import logging

import requests

from panoramic.cli.errors import (
    DatasetWriteException,
    FieldWriteException,
    InvalidDatasetException,
    InvalidFieldException,
    InvalidModelException,
    ModelWriteException,
)
from panoramic.cli.field import FieldClient
from panoramic.cli.field_mapper import map_field_from_local
from panoramic.cli.model import ModelClient
from panoramic.cli.model_mapper import map_data_source_from_local, map_model_from_local
from panoramic.cli.pano_model import (
    Actionable,
    PanoField,
    PanoModel,
    PanoVirtualDataSource,
)
from panoramic.cli.virtual_data_source import VirtualDataSourceClient

logger = logging.getLogger(__name__)


class RemoteWriter:

    """Responsible for writing data to remote API."""

    company_name: str
    virtual_data_source_client: VirtualDataSourceClient
    model_client: ModelClient
    field_client: FieldClient

    def __init__(self, company_name: str):
        self.company_name = company_name
        self.virtual_data_source_client = VirtualDataSourceClient()
        self.model_client = ModelClient()
        self.field_client = FieldClient()

    def delete(self, actionable: Actionable):
        """Delete data from remote api."""
        if isinstance(actionable, PanoModel):
            return self.delete_model(actionable)
        elif isinstance(actionable, PanoVirtualDataSource):
            return self.delete_data_source(actionable)
        elif isinstance(actionable, PanoField):
            return self.delete_field(actionable)
        else:
            raise NotImplementedError(f'delete not implemented for type {type(actionable)}')

    def write(self, actionable: Actionable):
        if isinstance(actionable, PanoModel):
            return self.write_model(actionable)
        elif isinstance(actionable, PanoVirtualDataSource):
            return self.write_data_source(actionable)
        elif isinstance(actionable, PanoField):
            return self.write_field(actionable)
        else:
            raise NotImplementedError(f'write not implemented for type {type(actionable)}')

    def write_data_source(self, data_source: PanoVirtualDataSource):
        """Write data source to remote API."""
        logger.debug(f'About to write data source {data_source.id}')
        remote_data_source = map_data_source_from_local(data_source)
        try:
            self.virtual_data_source_client.upsert_virtual_data_source(self.company_name, remote_data_source)
        except requests.RequestException as e:
            logger.debug(f'Failed to write dataset {data_source.dataset_slug}', exc_info=True)
            if e.response is not None and e.response.status_code == requests.codes.bad_request:
                raise InvalidDatasetException(e).extract_request_id(e)
            raise DatasetWriteException(data_source.dataset_slug).extract_request_id(e)

    def delete_data_source(self, data_source: PanoVirtualDataSource):
        """Delete data source from remote API."""
        logger.debug(f'About to delete data source {data_source.id}')
        try:
            self.virtual_data_source_client.delete_virtual_data_source(self.company_name, data_source.dataset_slug)
        except requests.RequestException as e:
            logger.debug(f'Failed to delete dataset {data_source.dataset_slug}', exc_info=True)
            raise DatasetWriteException(data_source.dataset_slug).extract_request_id(e)

    def write_model(self, model: PanoModel):
        """Write model to remote API."""
        logger.debug(f'About to write model {model.id}')
        # TODO: make virtual_data_source non optional
        assert model.virtual_data_source is not None
        remote_model = map_model_from_local(model)
        try:
            self.model_client.upsert_model(model.virtual_data_source, self.company_name, remote_model)
        except requests.RequestException as e:
            logger.debug(
                f'Failed to write model {model.model_name} in dataset {model.virtual_data_source}', exc_info=True
            )
            if e.response is not None and e.response.status_code == requests.codes.bad_request:
                raise InvalidModelException(e).extract_request_id(e)
            raise ModelWriteException(model.virtual_data_source, model.model_name).extract_request_id(e)

    def delete_model(self, model: PanoModel):
        """Delete model from remote API."""
        logger.debug(f'About to delete model {model.id}')
        # TODO: make virtual_data_source non optional
        assert model.virtual_data_source is not None
        try:
            self.model_client.delete_model(model.virtual_data_source, self.company_name, model.model_name)
        except requests.RequestException as e:
            logger.debug(
                f'Failed to delete model {model.model_name} in dataset {model.virtual_data_source}', exc_info=True
            )
            raise ModelWriteException(model.virtual_data_source, model.model_name).extract_request_id(e)

    def write_field(self, field: PanoField):
        """Write model on remote API."""
        logger.debug(f'About to write field {field.slug} in dataset {field.data_source}')

        remote_field = map_field_from_local(field)
        try:
            self.field_client.upsert_fields(company_slug=self.company_name, fields=[remote_field])
        except requests.RequestException as e:
            logger.debug(f'Failed to write field {field.slug} in dataset {field.data_source}', exc_info=True)
            if e.response is not None and e.response.status_code == requests.codes.bad_request:
                raise InvalidFieldException(e).extract_request_id(e)
            raise FieldWriteException(field.data_source, field.slug).extract_request_id(e)

    def delete_field(self, field: PanoField):
        """Delete model from remote API."""
        logger.debug(f'About to delete field {field.slug}')

        try:
            remote_field = map_field_from_local(field)
            self.field_client.delete_fields(company_slug=self.company_name, slugs=[remote_field.slug])
        except requests.RequestException as e:
            logger.debug(f'Failed to delete field {field.slug} in dataset {field.data_source}', exc_info=True)
            raise FieldWriteException(field.data_source, field.slug).extract_request_id(e)



import pytest
import responses

from panoramic.cli.model.client import Model, ModelClient


@pytest.fixture(autouse=True)
def mock_token_url(monkeypatch):
    monkeypatch.setenv('PANORAMIC_AUTH_TOKEN_URL', 'https://token')


@responses.activate
def test_delete_model():
    responses.add(responses.POST, 'https://token/', json={'access_token': '123123'})
    responses.add(responses.DELETE, 'https://diesel/model?virtual_data_source=test-source&company_slug=test-company')

    client = ModelClient(base_url='https://diesel/', client_id='client-id', client_secret='client-secret')

    client.delete_model('test-source', 'test-company', 'model')

    assert len(responses.calls) == 2


@responses.activate
def test_get_models():
    responses.add(responses.POST, 'https://token/', json={'access_token': '123123'})
    model = Model(
        model_name='model', data_source="db.schema.table", fields=[], joins=[], identifiers=[], visibility='available'
    )
    responses.add(responses.GET, 'https://diesel/', json={'data': [model.to_dict()]})

    client = ModelClient(base_url='https://diesel/', client_id='client-id', client_secret='client-secret')

    assert client.get_models('test-source', 'test-company') == [model]


@responses.activate
def test_get_model_names():
    responses.add(responses.POST, 'https://token/', json={'access_token': '123123'})
    responses.add(responses.GET, 'https://diesel/model-name', json={'data': ['a', 'b']})

    client = ModelClient(base_url='https://diesel/', client_id='client-id', client_secret='client-secret')

    assert client.get_model_names('test-source', 'test-company') == ['a', 'b']


@responses.activate
def test_upsert_model():
    responses.add(responses.POST, 'https://token/', json={'access_token': '123123'})
    responses.add(
        responses.PUT,
        'https://diesel/?virtual_data_source=test-source&company_slug=test-company',
        json={},
    )

    client = ModelClient(base_url='https://diesel/', client_id='client-id', client_secret='client-secret')

    client.upsert_model(
        'test-source',
        'test-company',
        Model(
            model_name='model',
            data_source="db.schema.table",
            fields=[],
            joins=[],
            identifiers=[],
            visibility='available',
        ),
    )

    assert len(responses.calls) == 2



import logging
from typing import List

import requests
from requests.exceptions import RequestException

from panoramic.cli.errors import IdentifierException, SourceNotFoundException
from panoramic.cli.identifier import IdentifierClient, JobState

logger = logging.getLogger(__name__)


class IdentifierGenerator:

    """Generate identifiers for a given source and table."""

    source_id: str
    client: IdentifierClient

    def __init__(self, company_slug: str, source_id: str, client: IdentifierClient = None):
        self.company_slug = company_slug
        self.source_id = source_id

        if client is None:
            client = IdentifierClient()

        self.client = client

    def fetch_token(self):
        self.client.fetch_token()

    def generate(self, table_name: str, timeout: int = 60) -> List[str]:
        """Generate identifiers for a given table."""
        logger.debug(f'Starting identifier generator job for table {table_name}')
        try:
            job_id = self.client.create_identifier_job(self.company_slug, self.source_id, table_name)
            logger.debug(f'Identifier generator job with id {job_id} started for table {table_name}')
        except RequestException as e:
            if e.response is not None and e.response.status_code == requests.codes.not_found:
                raise SourceNotFoundException(self.source_id)
            raise IdentifierException(self.source_id, table_name).extract_request_id(e)

        try:
            state = self.client.wait_for_terminal_state(job_id, timeout=timeout)
            if state != JobState.COMPLETED:
                raise IdentifierException(self.source_id, table_name)

            logger.debug(f'Identifier generator job with id {job_id} completed for table {table_name}')
            return self.client.get_job_results(job_id)['identifiers']
        except RequestException as e:
            raise IdentifierException(self.source_id, table_name).extract_request_id(e)



import pytest
from requests.exceptions import RequestException
from requests.models import Response

from panoramic.cli.errors import RefreshException


def test_no_response_request_exception():
    with pytest.raises(
        RefreshException, match='^Metadata could not be refreshed for table some_table in data connection some_source$'
    ):
        e = RequestException('Failed to connect')
        raise RefreshException('some_source', 'some_table').extract_request_id(e)


def test_no_headers_request_exception():
    with pytest.raises(
        RefreshException, match='^Metadata could not be refreshed for table some_table in data connection some_source$'
    ):
        e = RequestException('Failed to connect', response={})
        raise RefreshException('some_source', 'some_table').extract_request_id(e)


def test_no_request_id_request_exception():
    with pytest.raises(
        RefreshException, match='^Metadata could not be refreshed for table some_table in data connection some_source$'
    ):
        e = RequestException('Failed to connect', response=Response())
        raise RefreshException('some_source', 'some_table').extract_request_id(e)


def test_valid_request_id_request_exception():
    with pytest.raises(
        RefreshException,
        match=r'^Metadata could not be refreshed for table some_table in data connection some_source \(RequestId\: some_request_id\)$',
    ):
        response = Response()
        response.headers['x-diesel-request-id'] = 'some_request_id'
        e = RequestException('Failed to connect', response=response)
        raise RefreshException('some_source', 'some_table').extract_request_id(e)



from pathlib import Path
from unittest.mock import call, patch

from panoramic.cli.dbt import prepare_dbt_project
from panoramic.cli.paths import Paths


@patch('panoramic.cli.dbt.get_dbt_vars')
@patch('panoramic.cli.dbt.get_dbt_target_parts')
@patch('panoramic.cli.dbt.get_dbt_target_name')
@patch('panoramic.cli.dbt.get_dbt_packages')
@patch('panoramic.cli.dbt.write_yaml')
@patch('panoramic.cli.dbt.Connections.load')
def test_prepare_dbt_project(
    mock_connections_load,
    mock_write_yaml,
    mock_get_dbt_packages,
    mock_get_dbt_target_name,
    mock_get_dbt_target_parts,
    mock_get_dbt_vars,
):
    mock_connections_load.return_value = {
        'default-target': {
            'user': 'test-user',
            'password': 'test-password',
            'host': 'test-host',
        }
    }
    mock_get_dbt_target_name.return_value = 'default-target'
    mock_get_dbt_target_parts.return_value = {
        'schema': 'test-schema',
        'database': 'test-database',
    }
    mock_get_dbt_packages.return_value = [
        {'local': 'tmp/package'},
        {'package': 'panoramichq/dbt-test', 'version': '0.1.0'},
        {'git': 'https://github.com/fishtown-analytics/dbt-utils.git', 'revision': 'master'},
    ]
    mock_get_dbt_vars.return_value = {'var1': 'val1'}

    prepare_dbt_project()

    assert mock_write_yaml.mock_calls == [
        call(
            Paths.dbt_profiles_file(),
            {
                'default': {
                    'outputs': {
                        'default-target': {
                            'user': 'test-user',
                            'password': 'test-password',
                            'host': 'test-host',
                            'schema': 'test-schema',
                            'database': 'test-database',
                        }
                    },
                    'target': 'default-target',
                }
            },
        ),
        call(
            Paths.dbt_packages_file(),
            {
                'packages': [
                    {'local': str(Path.cwd() / 'tmp/package')},
                    {'package': 'panoramichq/dbt-test', 'version': '0.1.0'},
                    {'git': 'https://github.com/fishtown-analytics/dbt-utils.git', 'revision': 'master'},
                ]
            },
        ),
        call(
            Paths.dbt_project_file(),
            {
                'profile': 'default',
                'name': 'pano_dbt_temp',
                'version': '1.0.0',
                'config-version': 2,
                'models': {'+materialized': 'view'},
                'vars': {'var1': 'val1'},
            },
        ),
    ]



from unittest.mock import Mock

import pytest

from panoramic.cli.state import Action, VirtualState


def test_state_empty():
    assert VirtualState(data_sources=[], models=[], fields=[]).is_empty


def test_state_not_empty():
    assert not VirtualState(data_sources=[Mock()], models=[], fields=[]).is_empty
    assert not VirtualState(data_sources=[], models=[Mock()], fields=[]).is_empty
    assert not VirtualState(data_sources=[], models=[Mock()], fields=[Mock()]).is_empty


def test_empty_action_is_invalid():
    with pytest.raises(ValueError, match='cannot be None'):
        Action()


def test_get_objects_by_package():
    package1_model = Mock(package='package1')
    package2_model = Mock(package='package2')
    global_field = Mock(package=None)
    package1_field = Mock(package='package1')
    package3_field = Mock(package='package3')
    state = VirtualState(
        data_sources=[],
        models=[package1_model, package2_model],
        fields=[global_field, package1_field, package3_field],
    )

    assert state.get_objects_by_package() == {
        'package1': ([package1_field], [package1_model]),
        'package2': ([], [package2_model]),
        'package3': ([package3_field], []),
    }



from unittest.mock import Mock

import pytest
from requests.exceptions import RequestException

from panoramic.cli.errors import IdentifierException
from panoramic.cli.identifier.client import JobState
from panoramic.cli.identifier_generator import IdentifierGenerator


def test_generate_success():
    mock_client = Mock()
    generator = IdentifierGenerator('test--company', 'test-source', client=mock_client)

    mock_client.create_identifier_job.return_value = 'test-job-id'
    mock_client.wait_for_terminal_state.return_value = JobState.COMPLETED
    mock_client.get_job_results.return_value = {'identifiers': ['a', 'b']}

    assert ['a', 'b'] == generator.generate('test-table')


def test_generate_job_creation_failed():
    mock_client = Mock()
    generator = IdentifierGenerator('test--company', 'test-source', client=mock_client)

    mock_client.create_identifier_job.side_effect = RequestException('test')

    with pytest.raises(IdentifierException):
        generator.generate('test-table')


def test_generate_job_failed():
    mock_client = Mock()
    generator = IdentifierGenerator('test--company', 'test-source', client=mock_client)

    mock_client.create_identifier_job.return_value = 'test-job-id'
    mock_client.wait_for_terminal_state.return_value = JobState.FAILED

    with pytest.raises(IdentifierException):
        generator.generate('test-table')


def test_generate_job_results_failed():
    mock_client = Mock()
    generator = IdentifierGenerator('test--company', 'test-source', client=mock_client)

    mock_client.create_identifier_job.return_value = 'test-job-id'
    mock_client.wait_for_terminal_state.return_value = JobState.COMPLETED
    mock_client.get_job_results.side_effect = RequestException('test')

    with pytest.raises(IdentifierException):
        generator.generate('test-table')



from typing import Optional

from panoramic.cli.executor import Executor
from panoramic.cli.local.writer import FileWriter
from panoramic.cli.state import Action


class LocalExecutor(Executor):

    """Executes actions against local filesystem."""

    writer: FileWriter

    def __init__(self, writer: Optional[FileWriter] = None):
        if writer is None:
            writer = FileWriter()
        self.writer = writer

    def _execute(self, action: Action):
        if action.is_creation:
            assert action.desired is not None
            self.writer.write(action.desired)
        elif action.is_deletion:
            assert action.current is not None
            self.writer.delete(action.current)
        else:
            # assume update
            assert action.desired is not None
            assert action.current is not None
            self.writer.write(action.desired, package=action.current.package, file_name=action.current.file_name)



import pytest
import responses

from panoramic.cli.identifier.client import TERMINAL_STATES, IdentifierClient, JobState


@pytest.fixture(autouse=True)
def mock_token_url(monkeypatch):
    monkeypatch.setenv('PANORAMIC_AUTH_TOKEN_URL', 'https://token')


@responses.activate
def test_create_identifier_job():
    responses.add(responses.POST, 'https://token/', json={'access_token': '123123'})
    responses.add(
        responses.POST,
        'https://diesel/test-source?company_slug=test-company&table_name=test-table',
        json={'data': {'job_id': 'test-job-id'}},
    )

    client = IdentifierClient(base_url='https://diesel/', client_id='client-id', client_secret='client-secret')

    assert client.create_identifier_job('test-company', 'test-source', 'test-table') == 'test-job-id'


@responses.activate
def test_get_job_status():
    responses.add(responses.POST, 'https://token/', json={'access_token': '123123'})
    responses.add(responses.GET, 'https://diesel/job/test-job-id', json={'data': {'status': 'RUNNING'}})

    client = IdentifierClient(base_url='https://diesel/', client_id='client-id', client_secret='client-secret')

    assert client.get_job_status('test-job-id') == JobState.RUNNING


@responses.activate
def test_get_job_results():
    responses.add(responses.POST, 'https://token/', json={'access_token': '123123'})
    responses.add(responses.GET, 'https://diesel/job/test-job-id', json={'data': ['a', 'b']})

    client = IdentifierClient(base_url='https://diesel/', client_id='client-id', client_secret='client-secret')

    assert client.get_job_results('test-job-id') == ['a', 'b']


@responses.activate
@pytest.mark.parametrize('final_state', TERMINAL_STATES)
def test_wait_for_terminal_state(final_state):
    responses.add(responses.POST, 'https://token/', json={'access_token': '123123'})
    responses.add(responses.GET, 'https://diesel/job/test-job-id', json={'data': {'status': 'RUNNING'}})
    responses.add(responses.GET, 'https://diesel/job/test-job-id', json={'data': {'status': final_state.value}})

    client = IdentifierClient(base_url='https://diesel/', client_id='client-id', client_secret='client-secret')

    assert client.wait_for_terminal_state('test-job-id') == final_state



from typing import Any, Dict, Optional, Tuple

from panoramic.cli.errors import MissingFieldFileError
from panoramic.cli.field.client import Aggregation, Field
from panoramic.cli.pano_model import PanoField

NAMESPACE_DELIMITER = '|'


def map_field_from_remote(field: Field) -> PanoField:
    """Convert a remote field to local field."""
    data_source, slug = map_field_slug_from_remote(field.slug)

    return PanoField(
        slug=slug,
        group=field.group,
        display_name=field.display_name,
        data_type=field.data_type,
        field_type=field.field_type,
        description=field.description,
        data_source=data_source,
        aggregation=field.aggregation,
        calculation=field.calculation,
        display_format=field.display_format,
    )


def map_field_slug_from_remote(slug: str) -> Tuple[Optional[str], str]:
    if NAMESPACE_DELIMITER in slug:
        data_source, mappped_slug = slug.split(NAMESPACE_DELIMITER, 1)
        return data_source, mappped_slug
    else:
        return None, slug


def map_field_slug_from_local(slug: str, data_source: Optional[str]) -> str:
    if data_source is not None:
        return data_source + NAMESPACE_DELIMITER + slug
    else:
        return slug


def map_field_from_local(field: PanoField) -> Field:
    """Convert a local field to a remote fields"""

    return Field(
        slug=map_field_slug_from_local(field.slug, field.data_source),
        group=field.group,
        display_name=field.display_name,
        data_type=field.data_type,
        field_type=field.field_type,
        description=field.description,
        data_source=field.data_source,
        aggregation=field.aggregation,
        calculation=field.calculation,
        display_format=field.display_format,
    )


def map_column_to_field(
    column: Dict[str, Any],
    slug: str,
    data_source: str,
    is_identifier: bool = False,
) -> PanoField:
    aggregation = (
        Aggregation(type=column['aggregation_type'], params=None)
        if column.get('aggregation_type') is not None
        else None
    )
    field_type = 'dimension' if is_identifier else column['taxon_type']

    return PanoField(
        slug=slug,
        field_type=field_type,
        data_source=data_source,
        display_name=slug,
        group='CLI',
        data_type=column['validation_type'],
        aggregation=aggregation,
    )


def map_error_to_field(error: MissingFieldFileError) -> PanoField:
    return PanoField(
        slug=error.field_slug,
        field_type='dimension' if error.identifier else 'TODO: add field_type',
        data_source=error.dataset_slug,
        display_name=error.field_slug,
        group='CLI',
        data_type='TODO: add data_type',
    )



from collections import defaultdict
from typing import Dict, Generic, List, Optional, Tuple, TypeVar

from panoramic.cli.pano_model import (
    Actionable,
    PanoField,
    PanoModel,
    PanoVirtualDataSource,
)

T = TypeVar('T', covariant=True, bound=Actionable)


class Action(Generic[T]):
    """Action that transitions state from current to desired."""

    current: Optional[T]
    desired: Optional[T]

    def __init__(self, *, current: Optional[T] = None, desired: Optional[T] = None):
        if current is None and desired is None:
            raise ValueError('Both current and desired cannot be None')

        self.current = current
        self.desired = desired

    @property
    def is_creation(self) -> bool:
        return self.current is None

    @property
    def is_deletion(self) -> bool:
        return self.desired is None

    @property
    def description(self) -> str:
        if self.is_deletion:
            assert self.current is not None
            return f'DELETE: {".".join(self.current.id)}'
        elif self.is_creation:
            assert self.desired is not None
            return f'CREATE: {".".join(self.desired.id)}'
        else:
            assert self.desired is not None
            return f'UPDATE: {".".join(self.desired.id)}'

    def __hash__(self) -> int:
        return hash(
            (
                self.current,
                self.desired,
            )
        )

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, Action):
            return False

        return (self.current, self.desired) == (o.current, o.desired)


class ActionList(Generic[T]):
    """Container for actions."""

    actions: List[Action[T]]

    def __init__(self, *, actions: Optional[List[Action[T]]] = None):
        self.actions = actions if actions is not None else []

    @property
    def is_empty(self) -> bool:
        return len(self.actions) == 0

    @property
    def count(self) -> int:
        return len(self.actions)

    def add_action(self, action: Action):
        self.actions.append(action)

    def __hash__(self) -> int:
        return hash(tuple(self.actions))

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, ActionList):
            return False

        return self.actions == o.actions


class VirtualState:
    """Represent collection of virtual data sources."""

    data_sources: List[PanoVirtualDataSource]
    models: List[PanoModel]
    fields: List[PanoField]

    def __init__(self, *, data_sources: List[PanoVirtualDataSource], models: List[PanoModel], fields: List[PanoField]):
        self.data_sources = data_sources
        self.models = models
        self.fields = fields

    @property
    def is_empty(self) -> bool:
        return len(self.data_sources) == 0 and len(self.models) == 0 and len(self.fields) == 0

    def get_objects_by_package(self) -> Dict[str, Tuple[List[PanoField], List[PanoModel]]]:
        objects_by_dataset: Dict[str, Tuple[List[PanoField], List[PanoModel]]] = defaultdict(lambda: ([], []))

        for field in self.fields:
            # exclude global fields
            if field.package is not None:
                objects_by_dataset[field.package][0].append(field)

        for model in self.models:
            # models always have a package
            assert model.package is not None
            objects_by_dataset[model.package][1].append(model)

        return dict(objects_by_dataset)



import functools
import itertools
import json
from collections import defaultdict
from concurrent.futures import ThreadPoolExecutor
from pathlib import Path
from typing import Any, Dict, List, Tuple, Union

import jsonschema
from jsonschema.exceptions import ValidationError as JsonSchemaValidationError

from panoramic.cli.config.auth import get_client_id_env_var, get_client_secret_env_var
from panoramic.cli.errors import (
    DeprecatedAttributeWarning,
    DeprecatedConfigProperty,
    DuplicateFieldSlugError,
    DuplicateModelNameError,
    InvalidYamlFile,
    JsonSchemaError,
    MissingFieldFileError,
    OrphanFieldFileError,
    ValidationError,
)
from panoramic.cli.file_utils import read_yaml
from panoramic.cli.local.reader import FilePackage, FileReader, GlobalPackage
from panoramic.cli.pano_model import PanoField, PanoModel
from panoramic.cli.paths import Paths
from panoramic.cli.print import echo_warning


class JsonSchemas:
    @staticmethod
    @functools.lru_cache()
    def dataset() -> Dict[str, Any]:
        """Return schema for dataset files."""
        with Paths.dataset_schema_file().open('r') as f:
            return json.load(f)

    @staticmethod
    @functools.lru_cache()
    def model() -> Dict[str, Any]:
        """Return schema of model files."""
        with Paths.model_schema_file().open('r') as f:
            return json.load(f)

    @staticmethod
    @functools.lru_cache()
    def field() -> Dict[str, Any]:
        """Return schema of model files."""
        with Paths.field_schema_file().open('r') as f:
            return json.load(f)

    @staticmethod
    @functools.lru_cache()
    def config() -> Dict[str, Any]:
        with Paths.config_schema_file().open('r') as f:
            return json.load(f)

    @staticmethod
    @functools.lru_cache()
    def context():
        """Return schema of context file."""
        with Paths.context_schema_file().open('r') as f:
            return json.load(f)


def _check_model_deprecations(data: Dict[str, Any], path: Path) -> List[DeprecatedAttributeWarning]:
    """Check for deprecated attributes in a model."""
    errors = []

    fields: List[Dict[str, Any]] = data.get('fields', [])
    contains_data_type = any('data_type' in f for f in fields)
    if contains_data_type:
        errors.append(DeprecatedAttributeWarning(attribute='data_type', path=path))

    return errors


def _check_properties_deprecations(fp: Path, schema: Dict[str, Any]):
    """Check for deprecated properties in config file.
    WARNING: this check currently doesn't support recursive lookup of properties!"""
    errors = []
    data = read_yaml(fp)
    for name, value in schema.get('properties', {}).items():
        if value.get("deprecated", False) and data.get(name) is not None:
            errors.append(DeprecatedConfigProperty(name, value.get("deprecationMessage")))
    return errors


def _validate_data(data: Dict[str, Any], schema: Dict[str, Any]):
    """Validate file against schema."""
    jsonschema.validate(data, schema)


def _validate_file(fp: Path, schema: Dict[str, Any]):
    """Validate file against schema."""
    try:
        data = read_yaml(fp)
        _validate_data(data, schema)
    except JsonSchemaValidationError as e:
        raise JsonSchemaError(path=fp, error=e)


def _validate_package_dataset(package: FilePackage) -> List[ValidationError]:
    """Validate all files in a given package."""
    errors: List[ValidationError] = []
    try:
        _validate_data(package.read_data_source(), JsonSchemas.dataset())
    except InvalidYamlFile as e:
        errors.append(e)
    except JsonSchemaValidationError as e:
        errors.append(JsonSchemaError(path=package.data_source_file, error=e))

    return errors


def _validate_package_models(package: FilePackage) -> Tuple[List[PanoModel], List[ValidationError]]:
    """Validate models in a given package."""
    errors: List[ValidationError] = []
    models = []
    model_paths_by_name: Dict[str, List[Path]] = defaultdict(list)
    for model_data, model_path in package.read_models():
        try:
            _validate_data(model_data, JsonSchemas.model())
            model = PanoModel.from_dict(model_data)
            models.append(model)
            model_paths_by_name[model.model_name].append(model_path)
            errors.extend(_check_model_deprecations(model_data, model_path))
        except InvalidYamlFile as e:
            errors.append(e)
        except JsonSchemaValidationError as e:
            errors.append(JsonSchemaError(path=model_path, error=e))

    # check for duplicate model names
    for model_name, paths in model_paths_by_name.items():
        if len(paths) > 1:
            errors.append(DuplicateModelNameError(model_name=model_name, paths=paths))

    return models, errors


def _validate_package_fields(
    package: Union[FilePackage, GlobalPackage]
) -> Tuple[List[PanoField], List[ValidationError]]:
    errors: List[ValidationError] = []
    fields = []
    field_paths_by_id: Dict[Tuple, List[Path]] = defaultdict(list)
    for field_data, field_path in package.read_fields():
        try:
            _validate_data(field_data, JsonSchemas.field())
            field = PanoField.from_dict(field_data)
            fields.append(field)
            field_paths_by_id[(field.data_source, field.slug)].append(field_path)
        except InvalidYamlFile as e:
            errors.append(e)
        except JsonSchemaValidationError as e:
            errors.append(JsonSchemaError(path=field_path, error=e))

    # check for duplicate field slugs
    for (dataset_slug, field_slug), paths in field_paths_by_id.items():
        if len(paths) > 1:
            errors.append(DuplicateFieldSlugError(field_slug=field_slug, dataset_slug=dataset_slug, paths=paths))

    return fields, errors


def validate_missing_files(
    fields: List[PanoField], models: List[PanoModel], package_name: str
) -> List[MissingFieldFileError]:
    """Check for missing field files based on field map in model files."""
    fields_slugs_from_fields = {f.slug for f in fields}
    # take one model for every field
    data_reference_by_field_slugs = {
        field_name: (model.data_source, f.data_reference, field_name in model.identifiers)
        for model in models
        for f in model.fields
        for field_name in f.field_map
    }

    field_slugs_with_no_files = data_reference_by_field_slugs.keys() - fields_slugs_from_fields

    return [
        MissingFieldFileError(
            field_slug=slug,
            dataset_slug=package_name,
            data_source=data_reference_by_field_slugs[slug][0],
            data_reference=data_reference_by_field_slugs[slug][1],
            identifier=data_reference_by_field_slugs[slug][2],
        )
        for slug in field_slugs_with_no_files
    ]


def validate_orphaned_files(
    fields: List[PanoField], models: List[PanoModel], package_name: str
) -> List[OrphanFieldFileError]:
    """Check for field files not linked to any models."""
    # Fields with calculation without model link are normal
    fields_slugs_from_fields = {f.slug for f in fields if f.calculation is None}
    fields_slugs_from_models = set(itertools.chain.from_iterable(f.field_map for model in models for f in model.fields))

    extraneous_field_slugs = fields_slugs_from_fields.difference(fields_slugs_from_models)

    return [OrphanFieldFileError(field_slug=slug, dataset_slug=package_name) for slug in extraneous_field_slugs]


def _validate_package(package: FilePackage) -> List[ValidationError]:
    """Validate all files in a given package."""
    errors: List[ValidationError] = []
    errors.extend(_validate_package_dataset(package))

    models, model_errors = _validate_package_models(package)
    errors.extend(model_errors)

    fields, field_errors = _validate_package_fields(package)
    errors.extend(field_errors)

    missing_file_errors = validate_missing_files(fields, models, package_name=package.name)
    errors.extend(missing_file_errors)

    orphan_file_errors = validate_orphaned_files(fields, models, package_name=package.name)
    errors.extend(orphan_file_errors)

    return errors


def validate_local_state() -> List[ValidationError]:
    """Check local state against defined schemas."""
    file_reader = FileReader()
    packages = file_reader.get_packages()
    _, errors = _validate_package_fields(file_reader.get_global_package())

    executor = ThreadPoolExecutor(max_workers=4)
    for package_errors in executor.map(_validate_package, packages):
        errors.extend(package_errors)

    return errors


def validate_config():
    """Check config file against schema."""
    try:
        path, schema = Paths.config_file(), JsonSchemas.config()
        _validate_file(path, schema)
        errors = _check_properties_deprecations(path, schema)
        for err in errors:
            echo_warning(str(err))
    except ValidationError as e:
        try:
            # Valid if we get both values from env vars
            get_client_id_env_var()
            get_client_secret_env_var()
        except Exception:
            # Raise original exception if env vars don't help
            raise e


def validate_context():
    """Check context file against schema."""
    _validate_file(Paths.context_file(), JsonSchemas.context())


def validate_dbt():
    """Check context and config files contain valid DBT configuration."""
    pass



import sys

import panoramic.cli.schemas

if sys.version_info >= (3, 7):
    from importlib import resources as importlib_resources
else:
    import importlib_resources as importlib_resources

from enum import Enum
from pathlib import Path


class Paths:
    @staticmethod
    def context_file() -> Path:
        return Path.cwd() / PresetFileName.CONTEXT.value

    @staticmethod
    def dbt_project_dir() -> Path:
        return Path.cwd() / SystemDirectory.DBT.value

    @classmethod
    def dbt_modules_dir(cls) -> Path:
        return cls.dbt_project_dir() / PresetFileName.DBT_MODULES_DIR.value

    @classmethod
    def dbt_packages_file(cls) -> Path:
        return cls.dbt_project_dir() / PresetFileName.DBT_PACKAGES_FILE.value

    @classmethod
    def dbt_project_file(cls) -> Path:
        return cls.dbt_project_dir() / PresetFileName.DBT_PROJECT_FILE.value

    @staticmethod
    def dotenv_file() -> Path:
        return Path.cwd() / PresetFileName.DOTENV.value

    @staticmethod
    def dbt_config_dir() -> Path:
        return Path.home() / PresetFileName.CONFIG_DIR.value / PresetFileName.DBT_DIR.value

    @classmethod
    def dbt_profiles_file(cls) -> Path:
        return cls.dbt_config_dir() / PresetFileName.DBT_PROFILES_FILE.value

    @staticmethod
    def config_dir() -> Path:
        return Path.home() / PresetFileName.CONFIG_DIR.value

    @classmethod
    def config_file(cls) -> Path:
        return cls.config_dir() / PresetFileName.CONFIG.value

    @staticmethod
    def scanned_dir() -> Path:
        return Path.cwd() / SystemDirectory.SCANNED.value

    @staticmethod
    def scanned_fields_dir() -> Path:
        return Paths.fields_dir(Paths.scanned_dir())

    @staticmethod
    def company_fields_dir() -> Path:
        return Path.cwd() / SystemDirectory.FIELDS.value

    @staticmethod
    def fields_dir(package: Path) -> Path:
        return package / SystemDirectory.FIELDS.value

    @staticmethod
    def dataset_schema_file() -> Path:
        with importlib_resources.path(panoramic.cli.schemas, PresetFileName.DATASET_SCHEMA.value) as path:
            return path

    @staticmethod
    def model_schema_file() -> Path:
        with importlib_resources.path(panoramic.cli.schemas, PresetFileName.MODEL_SCHEMA.value) as path:
            return path

    @staticmethod
    def field_schema_file() -> Path:
        with importlib_resources.path(panoramic.cli.schemas, PresetFileName.FIELD_SCHEMA.value) as path:
            return path

    @staticmethod
    def config_schema_file() -> Path:
        with importlib_resources.path(panoramic.cli.schemas, PresetFileName.CONFIG_SCHEMA.value) as path:
            return path

    @staticmethod
    def context_schema_file() -> Path:
        with importlib_resources.path(panoramic.cli.schemas, PresetFileName.CONTEXT_SCHEMA.value) as path:
            return path

    @classmethod
    def analytics_dir(cls) -> Path:
        return cls.config_dir() / PresetFileName.ANALYTICS_DIR.value

    @classmethod
    def analytics_events_file(cls) -> Path:
        return cls.analytics_dir() / PresetFileName.ANALYTICS_EVENTS.value

    @classmethod
    def analytics_metadata_file(cls) -> Path:
        return cls.analytics_dir() / PresetFileName.ANALYTICS_METADATA.value


class FileExtension(Enum):
    """
    Enumeration with all available file extensions
    """

    MODEL_YAML = '.model.yaml'
    FIELD_YAML = '.field.yaml'


class PresetFileName(Enum):
    """Enumeration with all available preset file names."""

    DATASET_YAML = 'dataset.yaml'
    CONTEXT = 'pano.yaml'
    DOTENV = '.env'
    CONFIG_DIR = '.pano'
    DBT_DIR = '.dbt'
    CONFIG = 'config'
    DBT_MODULES_DIR = 'dbt_modules'
    MODEL_SCHEMA = 'model.schema.json'
    FIELD_SCHEMA = 'field.schema.json'
    DATASET_SCHEMA = 'dataset.schema.json'
    CONFIG_SCHEMA = 'config.schema.json'
    CONTEXT_SCHEMA = 'context.schema.json'
    ANALYTICS_DIR = 'analytics'
    ANALYTICS_EVENTS = 'events.jsonl'
    ANALYTICS_METADATA = 'metadata.yaml'
    DBT_PROJECT_FILE = 'dbt_project.yml'
    DBT_PACKAGES_FILE = 'packages.yml'
    DBT_PROFILES_FILE = 'profiles.yml'


class SystemDirectory(Enum):
    SCANNED = 'scanned'
    FIELDS = 'fields'
    DBT = '.dbt'



import pytest
import responses

from panoramic.cli.field.client import Field, FieldClient


@pytest.fixture(autouse=True)
def mock_token_url(monkeypatch):
    monkeypatch.setenv('PANORAMIC_AUTH_TOKEN_URL', 'https://token')


def _create_field(**kwargs) -> Field:
    base = dict(
        slug='some_field', field_type='Namespaced', display_name='Some field', group='group', data_type='data_type'
    )
    base.update(kwargs)

    return Field.from_dict(base)


@responses.activate
def test_upsert_field():
    responses.add(responses.POST, 'https://token/', json={'access_token': '123123'})
    field = _create_field()
    client = FieldClient(base_url='https://diesel/field/', client_id='client-id', client_secret='client-secret')

    # Create with a data source
    responses.add(
        responses.POST,
        'https://diesel/field?company_slug=test-company',
        json=[field.to_dict()],
    )
    client.upsert_fields(company_slug='test-company', fields=[field])
    assert len(responses.calls) == 2


@responses.activate
def test_get_fields_dataset_scoped():
    responses.add(responses.POST, 'https://token/', json={'access_token': '123123'})
    field = _create_field()

    responses.add(
        responses.GET,
        'https://diesel/field?company_slug=test-company&virtual_data_source=test-source&offset=0&limit=100',
        json={'data': [field.to_dict()]},
    )

    client = FieldClient(base_url='https://diesel/field/', client_id='client-id', client_secret='client-secret')

    assert client.get_fields(data_source='test-source', company_slug='test-company') == [field]
    assert len(responses.calls) == 2


@responses.activate
def test_get_all_fields():
    responses.add(responses.POST, 'https://token/', json={'access_token': '123123'})
    field = _create_field()

    responses.add(
        responses.GET,
        'https://diesel/field?company_slug=test-company&offset=0&limit=100',
        json={'data': [field.to_dict()]},
    )

    client = FieldClient(base_url='https://diesel/field/', client_id='client-id', client_secret='client-secret')

    assert client.get_fields(company_slug='test-company') == [field]
    assert len(responses.calls) == 2


@responses.activate
def test_delete_field():
    responses.add(responses.POST, 'https://token/', json={'access_token': '123123'})
    responses.add(
        responses.POST,
        'https://diesel/field/delete?company_slug=test-company',
        json=['field', 'vds|another_field'],
    )

    client = FieldClient(
        base_url='https://diesel/field/',
        client_id='client-id',
        client_secret='client-secret',
    )

    client.delete_fields(company_slug='test-company', slugs=['field', 'vds|another_field'])
    assert len(responses.calls) == 2



import functools
from typing import Any, Dict, List, Optional

from panoramic.cli.file_utils import read_yaml
from panoramic.cli.paths import Paths


@functools.lru_cache()
def get_company_slug() -> str:
    """Return company slug from context."""
    return read_yaml(Paths.context_file())['company_slug']


@functools.lru_cache()
def get_dbt_packages() -> Optional[List[Dict[str, Any]]]:
    """Return dbt packages from context."""
    return read_yaml(Paths.context_file()).get('dbt', {}).get('packages')


@functools.lru_cache()
def get_dbt_vars() -> Optional[Dict[str, Any]]:
    """Return dbt variables from context."""
    return read_yaml(Paths.context_file()).get('dbt', {}).get('vars')


@functools.lru_cache()
def get_dbt_target_name() -> str:
    """Return dbt target from context."""
    # target name is required if DBT is set
    return read_yaml(Paths.context_file())['dbt']['target']['name']


@functools.lru_cache()
def get_dbt_target_parts() -> Dict[str, Any]:
    """Return dbt target parts from context."""
    return read_yaml(Paths.context_file())['dbt']['target']['parts']



import logging

import requests
from requests.exceptions import RequestException

from panoramic.cli.errors import RefreshException, SourceNotFoundException
from panoramic.cli.metadata import JobState, MetadataClient

logger = logging.getLogger(__name__)


class Refresher:

    """Scans metadata for a given source and scope."""

    source_id: str
    client: MetadataClient

    def __init__(self, company_slug: str, source_id: str, client: MetadataClient = None):
        self.company_slug = company_slug
        self.source_id = source_id

        if client is None:
            self.client = MetadataClient()

    def fetch_token(self):
        self.client.fetch_token()

    def refresh_table(self, table_name: str, timeout: int = 60):
        """Scan columns for a given source and scope."""
        logger.debug(f'Starting refresh metadata job for table {table_name}')
        try:
            job_id = self.client.create_refresh_job(self.company_slug, self.source_id, table_name)
            logger.debug(f'Refresh metadata job with id {job_id} started for table {table_name}')
        except RequestException as e:
            if e.response is not None and e.response.status_code == requests.codes.not_found:
                raise SourceNotFoundException(self.source_id).extract_request_id(e)
            raise RefreshException(self.source_id, table_name).extract_request_id(e)

        try:
            state = self.client.wait_for_terminal_state(job_id, timeout=timeout)
            if state != JobState.COMPLETED:
                raise RefreshException(self.source_id, table_name)

            logger.debug(f'Refresh metadata job with id {job_id} completed for table {table_name}')
        except RequestException as e:
            raise RefreshException(self.source_id, table_name).extract_request_id(e)



from pathlib import Path

from panoramic.cli.connections import Connections
from panoramic.cli.context import (
    get_dbt_packages,
    get_dbt_target_name,
    get_dbt_target_parts,
    get_dbt_vars,
)
from panoramic.cli.file_utils import write_yaml
from panoramic.cli.paths import Paths


def prepare_dbt_project():
    """Set up DBT project for DBT CLI execution."""

    # Find the target connection
    target_name = get_dbt_target_name()
    target_parts = get_dbt_target_parts()

    # Create the config file
    connections = Connections.load()
    for connection_name, connection_data in connections.items():
        if connection_name == target_name:
            connections[connection_name] = {**connection_data, **target_parts}

    write_yaml(Paths.dbt_profiles_file(), {'default': {'outputs': connections, 'target': target_name}})

    # Create the package file
    packages = get_dbt_packages()

    # convert local package paths to be absolute
    for package in packages:
        if 'local' in package:
            package['local'] = str(Path(package['local']).absolute())

    if packages is not None:
        write_yaml(Paths.dbt_packages_file(), {'packages': packages})

    # Create the project file
    project_data = {
        # Default values used in dbt_project.yml
        'profile': 'default',
        'name': 'pano_dbt_temp',
        'version': '1.0.0',
        'config-version': 2,
        # Use SQL views for all projections
        'models': {'+materialized': 'view'},
    }

    dbt_vars = get_dbt_vars()
    if dbt_vars is not None:
        project_data['vars'] = dbt_vars

    write_yaml(Paths.dbt_project_file(), project_data)



import logging
import os
import sys


def configure_logging():
    log_level = os.environ.get('LOG_LEVEL', 'ERROR').upper()
    logging.basicConfig(stream=sys.stdout, level=log_level)

    # Hide auth logs from output
    logging.getLogger('requests_oauthlib').setLevel(logging.INFO)



import logging
from typing import Any, Dict, List, Optional
from urllib.parse import urljoin

from panoramic.auth import OAuth2Client
from panoramic.cli.clients import VersionedClient
from panoramic.cli.config.auth import get_client_id, get_client_secret
from panoramic.cli.config.virtual_data_source import get_base_url

logger = logging.getLogger(__name__)


class VirtualDataSource:

    slug: str
    display_name: str

    def __init__(self, *, slug: str, display_name: str):
        self.slug = slug
        self.display_name = display_name

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> 'VirtualDataSource':
        return VirtualDataSource(slug=data['slug'], display_name=data['display_name'])

    def to_dict(self) -> Dict[str, Any]:
        return {
            'slug': self.slug,
            'display_name': self.display_name,
        }

    def __hash__(self) -> int:
        return hash(
            (
                self.slug,
                self.display_name,
            )
        )

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, VirtualDataSource):
            return False

        return self.to_dict() == o.to_dict()


class VirtualDataSourceClient(OAuth2Client, VersionedClient):

    """Metadata HTTP API client."""

    base_url: str
    _base_url_with_trailing_slash: str

    def __init__(
        self,
        base_url: Optional[str] = None,
        client_id: Optional[str] = None,
        client_secret: Optional[str] = None,
    ):
        client_id = client_id if client_id is not None else get_client_id()
        client_secret = client_secret if client_secret is not None else get_client_secret()

        # Since we need to request api/virtual?company_id=x and api/virtual/slug?company_id=1
        # the base gets corrected to not include trailing slash
        #
        # Check https://stackoverflow.com/questions/10893374/python-confusions-with-urljoin for more context
        self.base_url = base_url if base_url is not None else get_base_url()
        if self.base_url[-1] == '/':
            self._base_url_with_trailing_slash = self.base_url
            self.base_url = self.base_url[0:-1]
        else:
            # base_url is in it's correct form - without trailing slash
            self._base_url_with_trailing_slash = self.base_url + '/'

        super().__init__(client_id, client_secret)

    def upsert_virtual_data_source(self, company_slug: str, payload: VirtualDataSource):
        """Create a virtual data source for a company"""
        logger.debug(f'Upserting virtual data source with payload {payload} under company {company_slug}')
        params = {'company_slug': company_slug}
        response = self.session.put(self.base_url, json=payload.to_dict(), params=params, timeout=30)
        response.raise_for_status()

    def get_virtual_data_source(self, company_slug: str, slug: str) -> VirtualDataSource:
        """Retrieve a virtual data source"""
        logger.debug(f'Retrieving a virtual data source with slug {slug} under company {company_slug}')
        url = urljoin(self._base_url_with_trailing_slash, slug)
        params = {'company_slug': company_slug}
        response = self.session.get(url, params=params, timeout=30)
        response.raise_for_status()
        return VirtualDataSource.from_dict(response.json()['data'])

    def get_all_virtual_data_sources(
        self, company_slug: str, *, offset: int = 0, limit: int = 100
    ) -> List[VirtualDataSource]:
        """Retrieve all virtual data sources under a company"""
        logger.debug(f'Retrieving all virtual data sources under company {company_slug}')
        params = {'company_slug': company_slug, 'offset': offset, 'limit': limit}
        response = self.session.get(self.base_url, params=params, timeout=30)
        response.raise_for_status()
        return [VirtualDataSource.from_dict(d) for d in response.json()['data']]

    def delete_virtual_data_source(self, company_slug: str, slug: str):
        """Delete a virtual data source"""
        logger.debug(f'Deleting virtual data source with slug {slug} under company {company_slug}')
        params = {'company_slug': company_slug}
        url = urljoin(self._base_url_with_trailing_slash, slug)
        response = self.session.delete(url, params=params, timeout=30)
        response.raise_for_status()



import difflib

from panoramic.cli.file_utils import dump_yaml
from panoramic.cli.print import Color, echo_info, echo_style
from panoramic.cli.state import ActionList

_LINE_START_TO_COLOR = {
    '+': Color.GREEN,
    '-': Color.RED,
    '@': Color.BLUE,
}


def echo_diff(actions: ActionList):
    for action in actions.actions:
        if action.is_deletion:
            echo_style(action.description, fg=Color.RED)
        elif action.is_creation:
            echo_style(action.description, fg=Color.GREEN)
        else:
            # Assumes update
            echo_style(action.description, fg=Color.YELLOW)

            current_yaml = dump_yaml(action.current.to_dict()) if action.current is not None else ''
            desired_yaml = dump_yaml(action.desired.to_dict()) if action.desired is not None else ''

            assert current_yaml is not None and desired_yaml is not None

            current_yaml_lines = current_yaml.splitlines(keepends=True)
            desired_yaml_lines = desired_yaml.splitlines(keepends=True)

            diff = difflib.unified_diff(current_yaml_lines, desired_yaml_lines, fromfile='current', tofile='desired')

            for line in diff:
                color = _LINE_START_TO_COLOR.get(line[0])
                echo_style(line, fg=color, nl=False)

        echo_info('')



from unittest.mock import Mock

import pytest
from requests.exceptions import RequestException

from panoramic.cli.errors import JoinException
from panoramic.cli.join.client import JobState, JoinClient
from panoramic.cli.join_detector import JoinDetector


def test_detect_joins_success():
    mock_client = Mock(JoinClient, autospec=True)
    detector = JoinDetector('test-company', client=mock_client)

    mock_client.create_join_detection_job.return_value = 'test-job-id'
    mock_client.wait_for_terminal_state.return_value = JobState.COMPLETED
    mock_client.get_job_results.return_value = {'joins': {}}

    assert {} == detector.detect('dataset')


def test_detect_joins_job_creation_failed():
    mock_client = Mock(JoinClient, autospec=True)
    detector = JoinDetector('test-company', client=mock_client)

    mock_client.create_join_detection_job.side_effect = RequestException('test')

    with pytest.raises(JoinException):
        detector.detect('dataset')


def test_detect_joins_job_failed():
    mock_client = Mock(JoinClient, autospec=True)
    detector = JoinDetector('test-company', client=mock_client)

    mock_client.create_join_detection_job.return_value = 'test-job-id'
    mock_client.wait_for_terminal_state.return_value = JobState.FAILED

    with pytest.raises(JoinException):
        detector.detect('dataset')


def test_generate_job_results_failed():
    mock_client = Mock(JoinClient, autospec=True)
    detector = JoinDetector('test--company', client=mock_client)

    mock_client.create_join_detection_job.return_value = 'test-job-id'
    mock_client.wait_for_terminal_state.return_value = JobState.COMPLETED
    mock_client.get_job_results.side_effect = RequestException('test')

    with pytest.raises(JoinException):
        detector.detect('dataset')



import functools
import os
from typing import Any, Dict, Optional

from panoramic.cli.config.storage import read_config


def get_client_id_env_var() -> str:
    return os.environ['PANO_CLIENT_ID']


def get_client_secret_env_var() -> str:
    return os.environ['PANO_CLIENT_SECRET']


@functools.lru_cache()
def get_client_id() -> str:
    try:
        return get_client_id_env_var()
    except KeyError:
        config = read_config()
        if 'auth' in config:
            return config['auth']['client_id']
        # Backwards compatibility for auth credentials
        return config['client_id']


@functools.lru_cache()
def get_client_secret() -> str:
    try:
        return get_client_secret_env_var()
    except KeyError:
        config = read_config()
        if 'auth' in config:
            return config['auth']['client_secret']
        # Backwards compatibility for auth credentials
        return config['client_secret']


@functools.lru_cache()
def get_dbt_profiles() -> Optional[Dict[str, Any]]:
    return read_config('dbt').get('profiles')



import traceback
from enum import Enum
from typing import Any, List, Optional

import click


class Color(Enum):

    RED = 'red'
    GREEN = 'green'
    YELLOW = 'yellow'
    BLUE = 'blue'


def echo_warning(msg: str):
    click.echo(f'Warning: {msg}')


def echo_error(msg: str, exc_info=False):
    click.echo(f'Error: {msg}')

    if exc_info:
        click.echo(traceback.format_exc())


def echo_info(msg: str):
    click.echo(msg)


def echo_style(msg: str, fg: Optional[Color] = None, nl: bool = True):
    """
    Print colored output.

    Should not be active while progress bar is visible.
    """
    fg_value = fg.value if fg is not None else None
    click.secho(msg, fg=fg_value, nl=nl)


def echo_warnings(errors: List[Any]):
    for error in errors:
        echo_info('')
        echo_warning(str(error))


def echo_errors(errors: List[Any]):
    for error in errors:
        echo_info('')  # visually separate errors
        echo_error(str(error))



import pytest

from panoramic.cli.errors import MissingFieldFileError
from panoramic.cli.field.client import Aggregation, Field
from panoramic.cli.field_mapper import (
    map_column_to_field,
    map_error_to_field,
    map_field_from_local,
    map_field_from_remote,
)
from panoramic.cli.pano_model import PanoField


@pytest.fixture()
def dummy_local_field() -> PanoField:
    data = dict(
        slug='slug',
        field_type='field_type',
        group='group',
        display_name='Display name',
        data_type='data_type',
        aggregation={'type': 'aggregation_type', 'params': {'a': 'b'}},
    )
    return PanoField.from_dict(data)


@pytest.fixture()
def dummy_remote_field() -> Field:
    data = dict(
        slug='slug',
        field_type='field_type',
        group='group',
        display_name='Display name',
        data_type='data_type',
        aggregation={'type': 'aggregation_type', 'params': {'a': 'b'}},
    )
    return Field.from_dict(data)


def test_field_local_to_remote(dummy_remote_field, dummy_local_field):
    assert map_field_from_local(dummy_local_field) == dummy_remote_field


def test_field_local_to_remote_data_source_bound(dummy_remote_field, dummy_local_field):
    dummy_remote_field.slug = 'data_source|slug'
    dummy_remote_field.data_source = 'data_source'
    dummy_local_field.data_source = 'data_source'

    assert map_field_from_local(dummy_local_field) == dummy_remote_field


def test_field_remote_to_local(dummy_remote_field, dummy_local_field):
    assert map_field_from_remote(dummy_remote_field) == dummy_local_field


def test_field_remote_to_local_data_source_bound(dummy_remote_field, dummy_local_field):
    dummy_remote_field.slug = 'data_source|slug'
    dummy_remote_field.data_source = 'data_source'
    dummy_local_field.data_source = 'data_source'

    assert map_field_from_remote(dummy_remote_field) == dummy_local_field


def test_map_column_to_field_basic():
    actual = map_column_to_field(
        {
            'column_name': 'slug',
            'data_type': 'TEXT',
            'taxon_type': 'dimension',
            'field_map': ['slug'],
            'data_reference': '"slug"',
            'aggregation_type': None,
            'validation_type': 'text',
        },
        slug='field_slug',
        data_source='data_source',
    )
    assert actual == PanoField(
        slug='field_slug',
        data_source='data_source',
        field_type='dimension',
        group='CLI',
        display_name='field_slug',
        data_type='text',
    )


def test_map_column_to_field_aggregation():
    actual = map_column_to_field(
        {
            'column_name': 'slug',
            'data_type': 'FLOAT',
            'taxon_type': 'metric',
            'field_map': ['slug'],
            'data_reference': '"slug"',
            'aggregation_type': 'sum',
            'validation_type': 'numeric',
        },
        slug='field_slug',
        data_source='data_source',
    )
    assert actual == PanoField(
        slug='field_slug',
        field_type='metric',
        group='CLI',
        data_source='data_source',
        display_name='field_slug',
        data_type='numeric',
        aggregation=Aggregation(type='sum', params=None),
    )


def test_map_column_to_field_identifier():
    actual = map_column_to_field(
        {
            'column_name': 'slug',
            'data_type': 'TEXT',
            'taxon_type': 'metric',
            'field_map': ['slug'],
            'data_reference': '"slug"',
            'aggregation_type': None,
            'validation_type': 'text',
        },
        slug='field_slug',
        data_source='data_source',
        is_identifier=True,
    )
    assert actual == PanoField(
        slug='field_slug',
        data_source='data_source',
        field_type='dimension',
        group='CLI',
        display_name='field_slug',
        data_type='text',
    )


def test_map_error_to_field():
    assert map_error_to_field(
        MissingFieldFileError(
            field_slug='test_field',
            dataset_slug='test_dataset',
            data_source='db.schema.table1',
            data_reference='"TEST_FIELD"',
            identifier=False,
        )
    ) == PanoField(
        slug='test_field',
        data_source='test_dataset',
        display_name='test_field',
        group='CLI',
        data_type='TODO: add data_type',
        field_type='TODO: add field_type',
    )



import itertools
import operator
from typing import Dict, Iterable

from panoramic.cli.model.client import Model, ModelField, ModelJoin
from panoramic.cli.pano_model import (
    PanoModel,
    PanoModelField,
    PanoModelJoin,
    PanoVirtualDataSource,
)
from panoramic.cli.virtual_data_source.client import VirtualDataSource


def map_data_source_from_remote(vds: VirtualDataSource) -> PanoVirtualDataSource:
    """Convert remote data source to local data source."""
    return PanoVirtualDataSource(dataset_slug=vds.slug, display_name=vds.display_name)


def map_data_source_from_local(vds: PanoVirtualDataSource) -> VirtualDataSource:
    """Convert local data source to remote data source."""
    return VirtualDataSource(slug=vds.dataset_slug, display_name=vds.display_name)


def map_model_join_from_remote(join: ModelJoin) -> PanoModelJoin:
    """Convert remote join to local join."""
    return PanoModelJoin(
        fields=join.fields, join_type=join.join_type, relationship=join.relationship, to_model=join.to_model
    )


def map_model_join_from_local(join: PanoModelJoin) -> ModelJoin:
    """Convert local join to remote join."""
    return ModelJoin(
        fields=join.fields, join_type=join.join_type, relationship=join.relationship, to_model=join.to_model
    )


def map_model_field_from_remote(field: ModelField) -> PanoModelField:
    """Convert remote attributes to local field."""
    return PanoModelField(field_map=field.field_map, data_reference=field.data_reference)


def map_attributes_from_local(field: PanoModelField) -> ModelField:
    """Convert local field to remote attributes."""
    return ModelField(
        data_reference=field.data_reference,
        field_map=field.field_map,
    )


def map_model_from_remote(model: Model) -> PanoModel:
    """Convert remote model to local model."""
    return PanoModel(
        model_name=model.model_name,
        data_source=model.data_source,
        fields=[map_model_field_from_remote(f) for f in model.fields],
        joins=[map_model_join_from_remote(j) for j in model.joins],
        identifiers=model.identifiers,
        virtual_data_source=model.virtual_data_source,
    )


def map_model_from_local(model: PanoModel) -> Model:
    """Convert local model to remote model."""
    return Model(
        model_name=model.model_name,
        data_source=model.data_source,
        fields=[map_attributes_from_local(f) for f in model.fields],
        joins=[map_model_join_from_local(j) for j in model.joins],
        identifiers=model.identifiers,
        visibility='available',  # default to available visibility
        virtual_data_source=model.virtual_data_source,
    )


def map_columns_to_model(raw_columns: Iterable[Dict]) -> Iterable[PanoModel]:
    """
    Load result of metadata table columns scan into Model
    """
    columns_grouped = itertools.groupby(raw_columns, operator.itemgetter('data_source', 'model_name'))

    for (data_source, model_name), columns in columns_grouped:
        yield PanoModel(
            model_name=model_name,
            data_source=data_source,
            fields=[
                PanoModelField(
                    data_reference=col['data_reference'],
                    field_map=col['field_map'],
                )
                for col in columns
            ],
            joins=[],
            identifiers=[],
        )



import functools
import os
import signal
import sys
from abc import ABC
from enum import Enum
from pathlib import Path
from typing import Callable, ClassVar, List, Optional

from jsonschema.exceptions import ValidationError as JsonSchemaValidationError
from requests.exceptions import RequestException
from yaml.error import MarkedYAMLError

from panoramic.cli.paths import Paths
from panoramic.cli.print import echo_error

DIESEL_REQUEST_ID_HEADER = 'x-diesel-request-id'


class CliBaseException(Exception):
    request_id: Optional[str] = None

    def add_request_id(self, request_id: str):
        self.request_id = request_id
        return self

    def extract_request_id(self, exc: RequestException):
        headers = getattr(exc.response, 'headers', {})
        return self.add_request_id(headers.get(DIESEL_REQUEST_ID_HEADER))

    def __str__(self) -> str:
        if self.request_id is not None:
            return f'{super().__str__()} (RequestId: {self.request_id})'
        return super().__str__()

    def __repr__(self) -> str:
        if self.request_id is not None:
            return f'{super().__repr__()} (RequestId: {self.request_id})'
        return super().__repr__()


class TimeoutException(CliBaseException):
    """Thrown when a remote operation times out."""


class IdentifierException(CliBaseException):
    """Error refreshing metadata."""

    def __init__(self, source_name: str, table_name: str):
        super().__init__(f'Identifiers could not be generated for table {table_name} in data connection {source_name}')


class JoinException(CliBaseException):
    """Error detecting joins in a dataset."""

    def __init__(self, dataset_name: str):
        super().__init__(f'Joins could not be detected for {dataset_name}')


class RefreshException(CliBaseException):
    """Error refreshing metadata."""

    def __init__(self, source_name: str, table_name: str):
        super().__init__(f'Metadata could not be refreshed for table {table_name} in data connection {source_name}')


class CompanyNotFoundException(CliBaseException):
    """Error when company not found for user (either no access or doesn't exist)."""

    def __init__(self, company_slug: str):
        super().__init__(f'Company {company_slug} not found. Do you have access to it?')


class SourceNotFoundException(CliBaseException):
    """Thrown when a source cannot be found."""

    def __init__(self, source_name: str):
        super().__init__(f'Data connection {source_name} not found. Has it been connected?')


class DatasetNotFoundException(CliBaseException):
    """Thrown when a dataset cannot be found."""

    def __init__(self, dataset_name: str):
        super().__init__(f'Dataset {dataset_name} not found. Has it been created?')


class ScanException(CliBaseException):
    """Error scanning metadata."""

    def __init__(self, source_name: str, table_filter: Optional[str]):
        table_msg = f' {table_filter} ' if table_filter is not None else ' '
        super().__init__(f'Metadata could not be scanned for table(s){table_msg}in data counnection: {source_name}')


class DatasetReadException(CliBaseException):
    """Error fetching virtual data sources."""

    def __init__(self, company_slug: str):
        super().__init__(f'Error fetching datasets for company {company_slug}')


class InvalidModelException(CliBaseException):
    """Invalid model submitted to remote."""

    messages: List[str]

    def __init__(self, error: RequestException):
        try:
            self.messages = [
                error['msg'] for error in error.response.json()['error']['extra_data']['validation_errors']
            ]
        except Exception:
            self.messages = ['Invalid model submitted']


class InvalidDatasetException(CliBaseException):
    """Invalid model submitted to remote."""

    messages: List[str]

    def __init__(self, error: RequestException):
        try:
            self.messages = [
                error['msg'] for error in error.response.json()['error']['extra_data']['validation_errors']
            ]
        except Exception:
            self.messages = ['Invalid dataset submitted']


class InvalidFieldException(CliBaseException):
    """Invalid field submitted to remote."""

    messages: List[str]

    def __init__(self, error: RequestException):
        try:
            self.messages = [
                error['msg'] for error in error.response.json()['error']['extra_data']['validation_errors']
            ]
        except Exception:
            self.messages = ['Invalid field submitted']


class DatasetWriteException(CliBaseException):
    """Error writing dataset to remote state."""

    def __init__(self, dataset_name: str):
        super().__init__(f'Error writing dataset {dataset_name}')


class ModelWriteException(CliBaseException):
    """Error writing dataset to remote state."""

    def __init__(self, dataset_name: str, model_name: str):
        super().__init__(f'Error writing model {model_name} in dataset {dataset_name}')


class ModelReadException(CliBaseException):
    """Error reading model(s) from remote state."""

    def __init__(self, company_slug: str, dataset_name: str):
        super().__init__(f'Error fetching models for company {company_slug} and dataset {dataset_name}')


class FieldWriteException(CliBaseException):
    """Error writing field to remote state."""

    def __init__(self, dataset_name: Optional[str], field_name: str):
        message = f'Error writing field {field_name}'
        if dataset_name is not None:
            message += f' in dataset {dataset_name}'

        super().__init__(message)


class FieldReadException(CliBaseException):
    """Error reading field(s) from remote state."""

    def __init__(self, company_slug: str, dataset_name: Optional[str]):
        dataset_message = f' under dataset {dataset_name}' if dataset_name else ''
        super().__init__(f'Error fetching field for company {company_slug}{dataset_message}')


class ValidationErrorSeverity(Enum):
    WARNING = 'WARNING'
    ERROR = 'ERROR'


class ValidationError(CliBaseException, ABC):
    """Abstract error raised during validation step."""

    severity: ClassVar[ValidationErrorSeverity] = ValidationErrorSeverity.ERROR


class FileMissingError(ValidationError):
    """File that should exist didn't."""

    def __init__(self, *, path: Path):
        if path == Paths.context_file():
            msg = f'Context file ({path.name}) not found in current working directory. Run pano init to create it.'
        elif path == Paths.config_file():
            msg = f'Config file ({path.absolute()}) not found. Run pano configure to create it.'
        else:
            # Should not happen => we only check above files exist explicitly
            msg = f'File Missing - {path}'

        super().__init__(msg)

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, FileMissingError):
            return False

        return str(self) == str(o)


class DuplicateModelNameError(ValidationError):
    """Two local models use the same model name."""

    def __init__(self, *, model_name: str, paths: List[Path]) -> None:
        try:
            paths = [path.relative_to(Path.cwd()) for path in paths]
        except ValueError:
            pass  # Use relative path when possible

        path_lines = ''.join(f'\n  in {path}' for path in paths)
        super().__init__(f'Multiple model files use model name {model_name}{path_lines}')

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, DuplicateModelNameError):
            return False

        return str(self) == str(o)


class DuplicateFieldSlugError(ValidationError):
    """Two local models use the same model name."""

    def __init__(self, *, field_slug: str, dataset_slug: Optional[str], paths: List[Path]) -> None:
        try:
            paths = [path.relative_to(Path.cwd()) for path in paths]
        except ValueError:
            pass  # Use relative path when possible

        path_lines = ''.join(f'\n  in {path}' for path in paths)
        dataset_message = f'under dataset {dataset_slug} ' if dataset_slug else ''
        super().__init__(f'Multiple field files {dataset_message}use slug {field_slug}{path_lines}')

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, DuplicateFieldSlugError):
            return False

        return str(self) == str(o)


class InvalidYamlFile(ValidationError):
    """YAML syntax error."""

    def __init__(self, *, path: Path, error: MarkedYAMLError):
        try:
            path = path.relative_to(Path.cwd())
        except ValueError:
            pass  # Use relative path when possible

        super().__init__(f'Invalid YAML file - {error.problem}\n  on line {error.problem_mark.line}\n  in {path}')

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, InvalidYamlFile):
            return False

        return str(self) == str(o)


class DeprecatedAttributeWarning(ValidationError):
    severity = ValidationErrorSeverity.WARNING

    def __init__(self, *, attribute: str, path: Path):
        try:
            path = path.relative_to(Path.cwd())
        except ValueError:
            pass  # Use relative path when possible

        super().__init__(f'Deprecated attribute "{attribute}" \n  in {path}')

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, DeprecatedAttributeWarning):
            return False

        return str(self) == str(o)


class DeprecatedConfigProperty(ValidationError):
    severity = ValidationErrorSeverity.WARNING

    def __init__(self, property_: str, deprecation_message: Optional[str] = None):
        if deprecation_message is None:
            deprecation_message = "Property is deprecated"
        super().__init__(f"'{property_}': {deprecation_message}\n  in {Paths.config_file()}")

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, DeprecatedConfigProperty):
            return False

        return str(self) == str(o)


class JsonSchemaError(ValidationError):
    def __init__(self, *, path: Path, error: JsonSchemaValidationError):
        try:
            path = path.relative_to(Path.cwd())
        except ValueError:
            pass  # Use relative path when possible

        error_path = '.'.join(str(p) for p in error.path)
        super().__init__(f'{error.message}\n  for path {error_path}\n  in {path}')

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, JsonSchemaError):
            return False

        return str(self) == str(o)


class OrphanFieldFileError(ValidationError):
    severity = ValidationErrorSeverity.WARNING
    field_slug: str
    dataset_slug: str

    def __init__(
        self,
        *,
        field_slug: str,
        dataset_slug: str,
    ) -> None:
        self.field_slug = field_slug
        self.dataset_slug = dataset_slug
        super().__init__(f'Field {field_slug} under dataset {dataset_slug} not used by any model')

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, OrphanFieldFileError):
            return False

        return str(self) == str(o)


class MissingFieldFileError(ValidationError):
    field_slug: str
    dataset_slug: str
    data_source: str
    data_reference: str
    identifier: bool

    def __init__(
        self,
        *,
        field_slug: str,
        dataset_slug: str,
        data_source: str,
        data_reference: str,
        identifier: bool,
    ) -> None:
        self.field_slug = field_slug
        self.dataset_slug = dataset_slug
        self.data_source = data_source
        self.data_reference = data_reference
        self.identifier = identifier
        super().__init__(f'Missing field file for slug {field_slug} under dataset {dataset_slug}')

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, MissingFieldFileError):
            return False

        return str(self) == str(o)


def handle_exception(f: Callable):
    """Print exception and exit with error code."""

    @functools.wraps(f)
    def wrapped(*args, **kwargs):
        try:
            return f(*args, **kwargs)
        except Exception:
            echo_error('Internal error occurred', exc_info=True)
            sys.exit(1)

    return wrapped


def handle_interrupt(f: Callable):
    """Exit app on keyboard interrupt."""

    @functools.wraps(f)
    def wrapped(*args, **kwargs):
        try:
            return f(*args, **kwargs)
        except KeyboardInterrupt:
            os._exit(128 + signal.SIGINT)

    return wrapped


class ConnectionNotFound(CliBaseException):
    """Connection not found in config."""

    def __init__(self, connection_name: str):
        super().__init__(f'Connection with name "{connection_name}" was not found.')


class ConnectionAlreadyExistsException(CliBaseException):
    """Connection with given name already exists."""

    def __init__(self, name: str):
        super().__init__(f'Connection with name "{name}" already exists.')


class ConnectionCreateException(CliBaseException):
    """Failed to create connection due to error."""

    def __init__(self, error_message: str):
        super().__init__(f'Failed to create connection: {error_message}.')


class ConnectionUpdateException(CliBaseException):
    """Failed to update connection due to error."""

    def __init__(self, error_message: str):
        super().__init__(f'Failed to update connection: {error_message}.')



from typing import Any, Dict

from panoramic.cli.file_utils import read_yaml, write_yaml
from panoramic.cli.paths import Paths


def update_config(section: str, data: Dict[str, Any]) -> None:
    data = {section: data}
    config_file = Paths.config_file()
    if config_file.is_file():
        config_yaml = read_yaml(config_file)
        config_yaml.update(data)
        write_yaml(config_file, config_yaml)
    else:
        write_yaml(config_file, data)


def read_config(section: str = '') -> Dict[str, Any]:
    config_file = Paths.config_file()
    if not config_file.is_file():
        return {}

    config = read_yaml(config_file)
    if section != '':
        return config.get(section, {})
    return config



import pytest
import responses

from panoramic.cli.companies.client import CompaniesClient


@pytest.fixture(autouse=True)
def mock_token_url(monkeypatch):
    monkeypatch.setenv('PANORAMIC_AUTH_TOKEN_URL', 'https://token')


@responses.activate
def test_get_companies():
    responses.add(responses.POST, 'https://token/', json={'access_token': '123123'})
    responses.add(
        responses.GET,
        'https://diesel/companies',
        json={'data': ['test-company-1', 'test-company-2', 'test-company-1']},
    )

    client = CompaniesClient(base_url='https://diesel/companies', client_id='client-id', client_secret='client-secret')

    assert client.get_companies() == {'test-company-1', 'test-company-2'}



from unittest.mock import Mock, patch

import pytest
from click.core import Context

from panoramic.cli.cli import (
    ConfigAwareCommand,
    ContextAwareCommand,
    LocalStateAwareCommand,
)
from panoramic.cli.errors import (
    CompanyNotFoundException,
    SourceNotFoundException,
    ValidationError,
)


@patch('panoramic.cli.validate.validate_config')
@patch('panoramic.cli.validate.validate_context', side_effect=SourceNotFoundException('test'))
def test_context_aware_command_source_not_found(_, __, capsys):
    """Check command fails when no context."""
    with pytest.raises(SystemExit):
        ContextAwareCommand(name='test-command').invoke(Mock())

    assert capsys.readouterr().out == 'Error: Data connection test not found. Has it been connected?\n'


@patch('panoramic.cli.validate.validate_config')
@patch('panoramic.cli.validate.validate_context', side_effect=CompanyNotFoundException('test'))
def test_context_aware_command_company_not_found(_, __, capsys):
    """Check command fails when no context."""
    with pytest.raises(SystemExit):
        ContextAwareCommand(name='test-command').invoke(Mock())

    assert capsys.readouterr().out == 'Error: Company test not found. Do you have access to it?\n'


@patch('panoramic.cli.validate.validate_config')
@patch('panoramic.cli.validate.validate_context', side_effect=ValidationError('test'))
def test_context_aware_command_invalid(_, __, capsys):
    """Check command fails when no context."""
    with pytest.raises(SystemExit):
        ContextAwareCommand(name='test-command').invoke(Mock())

    assert capsys.readouterr().out == 'Error: test\n'


@patch('panoramic.cli.command.analytics_module_is_enabled', return_value=False)
@patch('panoramic.cli.analytics.is_enabled', return_value=False)
@patch('panoramic.cli.validate.validate_config')
@patch('panoramic.cli.validate.validate_context')
def test_context_aware_command_valid(_, __, ___, ____):
    """Check command succeeds when context exists."""

    def test_callback():
        return 10

    command = ContextAwareCommand(name='test-command', callback=test_callback)
    context = Context(command)

    assert command.invoke(context) == 10


@patch('panoramic.cli.validate.validate_config', side_effect=ValidationError('test'))
def test_config_aware_command_invalid(_, capsys):
    """Check command fails when no context."""
    with pytest.raises(SystemExit):
        ConfigAwareCommand(name='test-command').invoke(Mock())

    assert capsys.readouterr().out == 'Error: test\n'


@patch('panoramic.cli.command.analytics_module_is_enabled', return_value=False)
@patch('panoramic.cli.analytics.is_enabled', return_value=False)
@patch('panoramic.cli.validate.validate_config')
def test_config_aware_command_config_exists(_, __, ___):
    """Check command succeeds when context exists."""

    def test_callback():
        return 10

    command = ConfigAwareCommand(name='test-command', callback=test_callback)
    context = Context(command)

    assert command.invoke(context) == 10


@patch('panoramic.cli.validate.validate_config')
@patch('panoramic.cli.validate.validate_context')
@patch('panoramic.cli.validate.validate_local_state', return_value=[ValidationError('test')])
def test_local_state_aware_command_invalid(_, __, ___, capsys):
    """Check command fails when no context."""
    with pytest.raises(SystemExit):
        LocalStateAwareCommand(name='test-command').invoke(Mock())

    assert capsys.readouterr().out == '\nError: test\n'


@patch('panoramic.cli.command.analytics_module_is_enabled', return_value=False)
@patch('panoramic.cli.analytics.is_enabled', return_value=False)
@patch('panoramic.cli.validate.validate_config')
@patch('panoramic.cli.validate.validate_context')
@patch('panoramic.cli.validate.validate_local_state')
def test_local_state_aware_command_valid(_, __, ___, ____, _____):
    """Check command succeeds when context exists."""

    def test_callback():
        return 10

    command = ConfigAwareCommand(name='test-command', callback=test_callback)
    context = Context(command)

    assert command.invoke(context) == 10



import logging
import os
import sys
import time
from collections import defaultdict
from pathlib import Path
from typing import Optional, Tuple

import click
from click.core import Command, Context
from dbt.main import main as dbt_main
from dotenv import load_dotenv

from panoramic.cli.__version__ import __version__
from panoramic.cli.analytics import is_enabled, write_command_event
from panoramic.cli.command import configure_anonymous_analytics
from panoramic.cli.errors import (
    CompanyNotFoundException,
    SourceNotFoundException,
    ValidationError,
    ValidationErrorSeverity,
    handle_exception,
    handle_interrupt,
)
from panoramic.cli.paths import Paths
from panoramic.cli.print import echo_error, echo_errors, echo_info, echo_warnings
from panoramic.cli.validate import validate_dbt

_PROFILES_DIR_ARG = '--profiles-dir'
_PROJECT_DIR_ARG = '--project-dir'


class CommandWithAnalytics(Command):
    """Execute command and write usage metrics to a file that is conditionally flushed to external system."""

    def invoke(self, ctx: Context):
        # This will make sure to display opt-out message for existing users that already run pano configure.
        if not is_enabled():
            configure_anonymous_analytics()

        # If command is a subcommand, combine its name with the command name.
        group = ''
        if ctx.parent is not None and ctx.parent.command != cli:  # type: ignore
            group = ctx.parent.command.name  # type: ignore

        start_time = time.time()
        try:
            result = super().invoke(ctx)
            write_command_event(self.name, group, start_time)
            return result
        except Exception as e:
            write_command_event(self.name, group, start_time, error=str(e))
            raise e


class ConfigAwareCommand(CommandWithAnalytics):
    """Perform config file validation before running command."""

    def invoke(self, ctx: Context):
        from panoramic.cli.validate import validate_config

        cwd = Path.cwd()

        try:
            validate_config()
            return super().invoke(ctx)
        except ValidationError as e:
            echo_error(str(e))
            sys.exit(1)
        finally:
            os.chdir(cwd)  # DBT sets cwd so we need to reset it


class ConnectionAwareCommand(ConfigAwareCommand):
    def invoke(self, ctx: Context):
        try:
            return super().invoke(ctx)
        except Exception as e:
            echo_error(str(e))
            sys.exit(1)


class ContextAwareCommand(ConfigAwareCommand):
    """
    Perform config and context file validation before running command.

    Failure scenarios we handle:
      * Invalid config/context/local state - ValidationError
      * Company not found for user - CompanyNotFoundException
      * Source not found for company - SourceNotFoundException

    Any other error - we show stack trace.
    """

    def invoke(self, ctx: Context):
        from panoramic.cli.validate import validate_context

        try:
            validate_context()
            return super().invoke(ctx)
        except (ValidationError, SourceNotFoundException, CompanyNotFoundException) as e:
            echo_error(str(e))
            sys.exit(1)


class DbtCommand(ContextAwareCommand):
    """Command proxied to DBT CLI tool."""

    def invoke(self, ctx: Context):
        from panoramic.cli.dbt import prepare_dbt_project

        try:
            validate_dbt()
            prepare_dbt_project()
            return super().invoke(ctx)
        except Exception as e:
            # TODO: Catch DBT exception here?
            echo_error(str(e))
            sys.exit(1)


class LocalStateAwareCommand(ContextAwareCommand):
    """Perform config, context, and local state files validation before running command."""

    def invoke(self, ctx: Context):
        from panoramic.cli.validate import validate_local_state

        errors_by_severity: defaultdict = defaultdict(list)
        for error in validate_local_state():
            errors_by_severity[error.severity].append(error)

        if len(errors_by_severity[ValidationErrorSeverity.WARNING]) > 0:
            echo_warnings(errors_by_severity[ValidationErrorSeverity.WARNING])
            echo_info('')

        if len(errors_by_severity[ValidationErrorSeverity.ERROR]) > 0:
            echo_errors(errors_by_severity[ValidationErrorSeverity.ERROR])
            sys.exit(1)

        return super().invoke(ctx)


@click.group(context_settings={'help_option_names': ["-h", "--help"]}, help='')
@click.option('--debug', is_flag=True, help='Enables debug mode')
@click.version_option(__version__)
@handle_exception
def cli(debug):
    """Run checks at the beginning of every command."""
    if debug:
        logger = logging.getLogger()
        logger.setLevel("DEBUG")

    load_dotenv(dotenv_path=Paths.dotenv_file())

    from panoramic.cli.supported_version import is_version_supported

    if not is_version_supported(__version__):
        sys.exit(1)


@cli.command(help='Scan models from source', cls=ContextAwareCommand)
@click.argument('source-id', type=str, required=True)
@click.option('--filter', '-f', type=str, help='Filter down what schemas to scan')
@click.option('--generate-identifiers', '-i', is_flag=True, help='Generate identifiers for models')
@click.option('--parallel', '-p', type=int, default=8, help='Parallelize metadata scan')
@handle_exception
@handle_interrupt
def scan(source_id: str, filter: Optional[str], parallel: int, generate_identifiers: bool):
    from panoramic.cli.command import scan as scan_command

    scan_command(source_id, filter, parallel, generate_identifiers)


@cli.command(help='Pull models from remote', cls=LocalStateAwareCommand)
@click.option('--yes', '-y', is_flag=True, default=False, help='Automatically confirm all actions')
@click.option('--target-dataset', '-t', type=str, help='Target a specific dataset')
@click.option('--diff', '-d', is_flag=True, help='Show the difference between local and remote state')
@handle_exception
def pull(yes: bool, target_dataset: str, diff: bool):
    from panoramic.cli.command import pull as pull_command

    pull_command(yes=yes, target_dataset=target_dataset, diff=diff)


@cli.command(help='Push models to remote', cls=LocalStateAwareCommand)
@click.option('--yes', '-y', is_flag=True, default=False, help='Automatically confirm all actions')
@click.option('--target-dataset', '-t', type=str, help='Target a specific dataset')
@click.option('--diff', '-d', is_flag=True, help='Show the difference between local and remote state')
@handle_exception
def push(yes: bool, target_dataset: str, diff: bool):
    from panoramic.cli.command import push as push_command

    push_command(yes=yes, target_dataset=target_dataset, diff=diff)


@cli.command(help='Configure pano CLI options')
@handle_exception
def configure():
    from panoramic.cli.command import configure as config_command

    config_command()


@cli.command(help='Initialize metadata repository', cls=ConfigAwareCommand)
@handle_exception
def init():
    from panoramic.cli.command import initialize

    initialize()


@cli.command(help='List available data connections', cls=ContextAwareCommand)
@handle_exception
def list_connections():
    from panoramic.cli.command import list_connections as list_connections_command

    list_connections_command()


@cli.command(help='List available companies', cls=ConfigAwareCommand)
@handle_exception
def list_companies():
    from panoramic.cli.command import list_companies as list_companies_command

    list_companies_command()


@cli.command(help='Validate local files', cls=CommandWithAnalytics)
@handle_exception
def validate():
    from panoramic.cli.command import validate as validate_command

    if not validate_command():
        sys.exit(1)


@cli.command(help='Detect joins under a dataset', cls=LocalStateAwareCommand)
@click.option('--target-dataset', '-t', type=str, help='Target a specific dataset')
@click.option('--yes', '-y', is_flag=True, default=False, help='Automatically confirm all actions')
@click.option('--diff', '-d', is_flag=True, help='Show the difference between local and detected joins')
@click.option(
    '--overwrite', is_flag=True, default=False, help='Overwrite joins on local model files by suggestions from remote'
)
@handle_exception
def detect_joins(target_dataset: str, yes: bool, diff: bool, overwrite: bool):
    from panoramic.cli.command import detect_joins as detect_joins_command

    detect_joins_command(target_dataset=target_dataset, diff=diff, overwrite=overwrite, yes=yes)


@cli.group(name='field')
def field_cli():
    """Commands on local field files."""


@field_cli.command(help='Clean up fields not linked to any model', cls=ContextAwareCommand)
@click.option('--target-dataset', '-t', type=str, help='Target a specific dataset')
@click.option('--yes', '-y', is_flag=True, default=False, help='Automatically confirm all actions')
@handle_exception
def cleanup(target_dataset: str, yes: bool):
    from panoramic.cli.command import delete_orphaned_fields as delete_orphans_command

    delete_orphans_command(target_dataset=target_dataset, yes=yes)


@field_cli.command(help='Scaffold fields defined in models', cls=ContextAwareCommand)
@click.option('--target-dataset', '-t', type=str, help='Target a specific dataset')
@click.option('--yes', '-y', is_flag=True, default=False, help='Automatically confirm all actions')
@handle_exception
def scaffold(target_dataset: str, yes: bool):
    from panoramic.cli.command import (
        scaffold_missing_fields as scaffold_missing_fields_command,
    )

    scaffold_missing_fields_command(target_dataset=target_dataset, yes=yes)


@cli.group()
def connection():
    """Connection subcommand for managing connections.

    \b
    All connections are stored in ~/.pano/config file.
    You can edit this file either manually or using provided commands.

    Expected YAML structure:

    \b
    name:
      type: postgres
      user: my_user
      password: <password>
      host: localhost
      port: 5432
      database: my_db

    Which effectively results in a connection string similar to: "postgres://my_user@<password>@localhost:5432/my_db".

    \b
    name: The name of the connection that will be used as reference to specify which connection to use.
    type: The type of data warehouse you are connecting to.
    user: The username used for database connection.
    password: The password used for database connection.
    host: The hostname (server) used for database connection.
    port: The port used for database connection.
    database: The database used for database connection.
    schema: The schema to build models into by default. Can be overridden by custom models.
    warehouse: The warehouse to use when building models. (Used by Snowflake)
    account: The account used for database connection. This will be something like cc123 or cc123.us-east-1 for your particular account. (Used by Snowflake)
    project: The project used for database connection. (Used by BigQuery)
    key_file: Keyfile path to Service Account JSON. (Used by BigQuery)
    """
    pass


@connection.command(cls=ConnectionAwareCommand)
@click.argument('name', type=str)
@click.option('--type', default=None, type=str, help='Type of the database. E.g. "postgres", "snowflake".')
@click.option('--user', default=None, type=str, help='Connection username.')
@click.option('--password', default=None, type=str, help='Connection password.')
@click.option('--password-stdin', default=False, is_flag=True, help='Read connection password from standard input.')
@click.option('--host', default=None, type=str, help='Connection hostname.')
@click.option('--port', default=None, type=int, help='Connection port.')
@click.option('--database', default=None, type=str, help='Connection database name.')
@click.option('--schema', default=None, type=str, help='Connection schema name.')
@click.option('--warehouse', default=None, type=str, help='Connection warehouse name. (Used by Snowflake)')
@click.option('--account', default=None, type=str, help='Connection account name. (Used by Snowflake)')
@click.option('--project', default=None, type=str, help='Connection project name. (Used by BigQuery)')
@click.option('--key-file', default=None, type=str, help='Keyfile path to Service Account JSON. (Used by BigQuery)')
@click.option('--no-test', default=False, is_flag=True, help='Do NOT try test the connection.')
def create(
    name: str,
    type: Optional[str],
    user: Optional[str],
    host: Optional[str],
    port: Optional[int],
    password: Optional[str],
    password_stdin: bool,
    database: Optional[str],
    schema: Optional[str],
    warehouse: Optional[str],
    account: Optional[str],
    project: Optional[str],
    key_file: Optional[str],
    no_test: bool,
):
    """Add new connection.

    pano connection create postgres-prod --type postgres --user my_user \\
     --password-stdin --host localhost --port 5432 --database my_db
    """
    from panoramic.cli.connections import create_connection_command

    create_connection_command(
        name=name,
        type=type,
        user=user,
        host=host,
        port=port,
        password=password,
        password_stdin=password_stdin,
        database=database,
        schema=schema,
        warehouse=warehouse,
        account=account,
        project=project,
        key_file=key_file,
        no_test=no_test,
    )


@connection.command(cls=ConnectionAwareCommand)
@click.argument('name', type=str)
@click.option('--type', default=None, type=str, help='Type of the database. E.g. "postgres", "snowflake".')
@click.option('--user', default=None, type=str, help='Connection username.')
@click.option('--password', default=None, type=str, help='Connection password.')
@click.option('--password-stdin', default=False, is_flag=True, help='Read connection password from standard input.')
@click.option('--host', default=None, type=str, help='Connection hostname.')
@click.option('--port', default=None, type=int, help='Connection port.')
@click.option('--database', default=None, type=str, help='Connection database name.')
@click.option('--schema', default=None, type=str, help='Connection schema name.')
@click.option('--warehouse', default=None, type=str, help='Connection warehouse name. (Used by Snowflake)')
@click.option('--account', default=None, type=str, help='Connection account name. (Used by Snowflake)')
@click.option('--project', default=None, type=str, help='Connection project name. (Used by BigQuery)')
@click.option('--key-file', default=None, type=str, help='Keyfile path to Service Account JSON. (Used by BigQuery)')
@click.option('--no-test', default=False, is_flag=True, help='Do NOT try test the connection.')
def update(
    name: str,
    type: Optional[str],
    user: Optional[str],
    host: Optional[str],
    port: Optional[int],
    password: Optional[str],
    password_stdin: bool,
    database: Optional[str],
    schema: Optional[str],
    warehouse: Optional[str],
    account: Optional[str],
    project: Optional[str],
    key_file: Optional[str],
    no_test: bool,
):
    """Update existing connection.

    pano connection update postgres-prod --database my_new_prod_db
    """
    from panoramic.cli.connections import update_connection_command

    update_connection_command(
        name=name,
        type=type,
        user=user,
        host=host,
        port=port,
        password=password,
        password_stdin=password_stdin,
        database=database,
        schema=schema,
        warehouse=warehouse,
        account=account,
        project=project,
        key_file=key_file,
        no_test=no_test,
    )


@connection.command(cls=ConnectionAwareCommand)
@click.argument('name', type=str)
def remove(name: str):
    """Remove existing connection.

    pano connection remove postgres-prod
    """
    from panoramic.cli.connections import remove_connection_command

    remove_connection_command(name)


@connection.command(name='list', cls=ConnectionAwareCommand)
@click.option('--show-password', default=False, is_flag=True, help='Show passwords.')
def list_(show_password: bool):  # we cannot have method 'list' due to conflicts
    """List all available connections.

    pano connection list --show-password
    """
    from panoramic.cli.connections import list_connections_command

    list_connections_command(show_password)


@connection.command(cls=ConnectionAwareCommand)
@click.argument('name', default=None, type=str, required=False)
def test(name: str):
    """Test connections.

    pano connection test
    """
    from panoramic.cli.connections import test_connections_command

    test_connections_command(name)


@cli.group()
def analytics():
    """Analytics subcommand for managing anonymous usage metrics collection preferences."""
    pass


@analytics.command(name='on', cls=ConfigAwareCommand)
def analytics_on():
    """Opt in to anonymous usage analytics.

    pano analytics on
    """
    from panoramic.cli.analytics import opt_in_command

    opt_in_command()


@analytics.command(name='off', cls=ConfigAwareCommand)
def analytics_off():
    """Opt out off anonymous usage analytics.

    pano analytics off
    """
    from panoramic.cli.analytics import opt_out_command

    opt_out_command()


@analytics.command(name='id', cls=ConfigAwareCommand)
def analytics_id():
    """Display current anonymous usage tracking id.

    pano analytics id
    """
    from panoramic.cli.analytics import show_tracking_id_command

    show_tracking_id_command()


@cli.command(context_settings=dict(ignore_unknown_options=True), cls=DbtCommand)
@click.argument('dbt_args', nargs=-1, type=click.UNPROCESSED)
def dbt(dbt_args: Tuple[str]):
    """Run DBT command.

    \b
    Supported commands:
      pano dbt deps     Update package dependencies from pano.yaml
      pano dbt compile  Compile pre-model transforms into .dbt/target/compiled
      pano dbt run      Run compiled transforms against target defined in pano.yaml
    """
    # Handle no args passed
    if len(dbt_args) == 0:
        ctx = click.get_current_context()
        click.echo(ctx.get_help())
        return

    args = list(dbt_args)

    if _PROFILES_DIR_ARG not in args:
        args.extend([_PROFILES_DIR_ARG, str(Paths.dbt_config_dir())])
    if _PROJECT_DIR_ARG not in args:
        args.extend([_PROJECT_DIR_ARG, str(Paths.dbt_project_dir())])

    return dbt_main(args=args)



import yaml

from panoramic.cli.context import get_company_slug
from panoramic.cli.paths import Paths


def test_context_file(monkeypatch, tmp_path):
    monkeypatch.chdir(tmp_path)

    with open(Paths.context_file(), 'w') as f:
        f.write(yaml.dump(dict(api_version='v2', company_slug='company_name_12fxs')))

    # Use __wrapped__ to avoid cache
    assert get_company_slug.__wrapped__() == 'company_name_12fxs'



from typing import Optional

from panoramic.cli.local.reader import FileReader
from panoramic.cli.pano_model import PanoField, PanoModel, PanoVirtualDataSource
from panoramic.cli.state import VirtualState


def get_state(target_dataset: Optional[str] = None) -> VirtualState:
    """
    Build a representation of what VDS and models are on local filesystem.
    """
    file_reader = FileReader()
    packages = file_reader.get_packages()
    data_sources = []
    models = []
    fields = []

    if target_dataset is None:
        for field, path in file_reader.get_global_package().read_fields():
            field['file_name'] = path.name
            fields.append(PanoField.from_dict(field))

    for package in packages:
        data_source = package.read_data_source()
        data_source['package'] = package.name
        pvds = PanoVirtualDataSource.from_dict(data_source)

        if target_dataset is not None and target_dataset != pvds.dataset_slug:
            continue

        data_sources.append(pvds)
        for model, path in package.read_models():
            model['package'] = package.name
            model['file_name'] = path.name
            model['virtual_data_source'] = pvds.dataset_slug
            models.append(PanoModel.from_dict(model))

        for field, path in package.read_fields():
            field['package'] = package.name
            field['file_name'] = path.name
            field['data_source'] = pvds.dataset_slug
            fields.append(PanoField.from_dict(field))

    return VirtualState(data_sources=data_sources, models=models, fields=fields)



import pytest
import responses

from panoramic.cli.metadata.client import TERMINAL_STATES, JobState, MetadataClient


@pytest.fixture(autouse=True)
def mock_token_url(monkeypatch):
    monkeypatch.setenv('PANORAMIC_AUTH_TOKEN_URL', 'https://token')


@responses.activate
def test_create_get_columns_job():
    responses.add(responses.POST, 'https://token/', json={'access_token': '123123'})
    responses.add(
        responses.POST,
        'https://diesel/test-source/columns?company_slug=test-company&table_filter=test-filter',
        json={'data': {'job_id': 'test-job-id'}},
    )

    client = MetadataClient(base_url='https://diesel/', client_id='client-id', client_secret='client-secret')

    assert client.create_get_columns_job('test-company', 'test-source', 'test-filter') == 'test-job-id'


@responses.activate
def test_create_get_tables_job():
    responses.add(responses.POST, 'https://token/', json={'access_token': '123123'})
    responses.add(
        responses.POST,
        'https://diesel/test-source/tables?company_slug=test-company&table_filter=test-filter',
        json={'data': {'job_id': 'test-job-id'}},
    )

    client = MetadataClient(base_url='https://diesel/', client_id='client-id', client_secret='client-secret')

    assert client.create_get_tables_job('test-company', 'test-source', 'test-filter') == 'test-job-id'


@responses.activate
def test_create_refresh_job():
    responses.add(responses.POST, 'https://token/', json={'access_token': '123123'})
    responses.add(
        responses.POST,
        'https://diesel/test-source/refresh?company_slug=test-company&table_name=test-table',
        json={'data': {'job_id': 'test-job-id'}},
    )

    client = MetadataClient(base_url='https://diesel/', client_id='client-id', client_secret='client-secret')

    assert client.create_refresh_job('test-company', 'test-source', 'test-table') == 'test-job-id'


@responses.activate
def test_get_job_status():
    responses.add(responses.POST, 'https://token/', json={'access_token': '123123'})
    responses.add(responses.GET, 'https://diesel/job/test-job-id', json={'data': {'job_status': 'RUNNING'}})

    client = MetadataClient(base_url='https://diesel/', client_id='client-id', client_secret='client-secret')

    assert client.get_job_status('test-job-id') == JobState.RUNNING


@responses.activate
def test_get_job_results():
    responses.add(responses.POST, 'https://token/', json={'access_token': '123123'})
    responses.add(responses.GET, 'https://diesel/job/test-job-id/results', json={'data': [{'a': 'b'}]})

    client = MetadataClient(base_url='https://diesel/', client_id='client-id', client_secret='client-secret')

    assert client.get_job_results('test-job-id') == [{'a': 'b'}]


@responses.activate
@pytest.mark.parametrize('final_state', TERMINAL_STATES)
def test_wait_for_terminal_state(final_state):
    responses.add(responses.POST, 'https://token/', json={'access_token': '123123'})
    responses.add(responses.GET, 'https://diesel/job/test-job-id', json={'data': {'job_status': 'RUNNING'}})
    responses.add(responses.GET, 'https://diesel/job/test-job-id', json={'data': {'job_status': final_state.value}})

    client = MetadataClient(base_url='https://diesel/', client_id='client-id', client_secret='client-secret')

    assert client.wait_for_terminal_state('test-job-id') == final_state


@responses.activate
def test_collect_results():
    responses.add(responses.POST, 'https://token/', json={'access_token': '123123'})
    responses.add(responses.GET, 'https://diesel/job/test-job-id/results?offset=0&limit=1', json={'data': [{'a': 'b'}]})
    responses.add(responses.GET, 'https://diesel/job/test-job-id/results?offset=1&limit=1', json={'data': [{'c': 'd'}]})
    responses.add(responses.GET, 'https://diesel/job/test-job-id/results?offset=2&limit=1', json={'data': []})

    client = MetadataClient(base_url='https://diesel/', client_id='client-id', client_secret='client-secret')

    assert list(client.collect_results('test-job-id', limit=1)) == [{'a': 'b'}, {'c': 'd'}]



import itertools
import logging
from collections import defaultdict
from typing import Dict, Iterable, List, Optional, Sequence, Tuple

import requests
from requests.exceptions import RequestException

from panoramic.cli.context import get_company_slug
from panoramic.cli.errors import (
    MissingFieldFileError,
    ScanException,
    SourceNotFoundException,
)
from panoramic.cli.field_mapper import map_column_to_field, map_error_to_field
from panoramic.cli.metadata import JobState, MetadataClient
from panoramic.cli.model_mapper import map_columns_to_model
from panoramic.cli.pano_model import PanoField, PanoModel

logger = logging.getLogger(__name__)


class Scanner:

    """Scans metadata for a given source and scope."""

    source_id: str
    client: MetadataClient

    def __init__(self, company_slug: str, source_id: str, client: MetadataClient = None):
        self.company_slug = company_slug
        self.source_id = source_id

        if client is None:
            self.client = MetadataClient()

    def fetch_token(self):
        self.client.fetch_token()

    def scan_tables(self, *, table_filter: Optional[str] = None, timeout: int = 60) -> Iterable[Dict]:
        """Scan tables for a given source and filter."""
        logger.debug(f'Starting get tables job with filter {table_filter}')
        try:
            job_id = self.client.create_get_tables_job(self.company_slug, self.source_id, table_filter)
            logger.debug(f'Get tables job with id {job_id} started with filter {table_filter}')
        except RequestException as e:
            if e.response is not None and e.response.status_code == requests.codes.not_found:
                raise SourceNotFoundException(self.source_id).extract_request_id(e)
            raise ScanException(self.source_id, table_filter).extract_request_id(e)

        try:
            state = self.client.wait_for_terminal_state(job_id, timeout=timeout)
            if state != JobState.COMPLETED:
                raise ScanException(self.source_id, table_filter)

            logger.debug(f'Get tables job with id {job_id} completed with filter {table_filter}')
            yield from self.client.collect_results(job_id)
        except RequestException as e:
            raise ScanException(self.source_id, table_filter).extract_request_id(e)

    def scan_columns(self, *, table_filter: Optional[str] = None, timeout: int = 60) -> Iterable[Dict]:
        """Scan columns for a given source and filter."""
        logger.debug('Starting get columns job')
        try:
            job_id = self.client.create_get_columns_job(self.company_slug, self.source_id, table_filter)
            logger.debug(f'Get columns job with id {job_id} started with filter {table_filter}')
        except requests.HTTPError as e:
            if e.response is not None and e.response.status_code == requests.codes.not_found:
                raise SourceNotFoundException(self.source_id).extract_request_id(e)
            raise ScanException(self.source_id, table_filter).extract_request_id(e)

        try:
            state = self.client.wait_for_terminal_state(job_id, timeout=timeout)
            if state != JobState.COMPLETED:
                raise ScanException(self.source_id, table_filter)

            logger.debug(f'Get columns job with id {job_id} completed with filter {table_filter}')
            yield from self.client.collect_results(job_id)
        except RequestException as e:
            raise ScanException(self.source_id, table_filter).extract_request_id(e)

    def scan_columns_grouped(self, *, table_filter: Optional[str] = None, timeout: int = 60) -> Iterable[PanoModel]:
        """Scan columns for a given source and group them by model and data source."""
        yield from map_columns_to_model(self.scan_columns(table_filter=table_filter, timeout=timeout))


def _group_errors_by_column(
    errors: Sequence[MissingFieldFileError],
) -> Dict[Tuple[str, str], List[MissingFieldFileError]]:
    """Group errors by data_source+data_reference (table+column)."""
    errors_by_column: Dict[Tuple[str, str], List[MissingFieldFileError]] = defaultdict(list)
    for error in errors:
        errors_by_column[(error.data_source, error.data_reference)].append(error)
    return dict(errors_by_column)


def scan_fields_for_errors(errors: Sequence[MissingFieldFileError]) -> List[PanoField]:
    """Scan fields for missing file errors."""
    company_slug = get_company_slug()

    fields: List[PanoField] = []
    errors_by_column = _group_errors_by_column(errors)
    data_sources = {data_source for data_source, _ in errors_by_column}

    for data_source in data_sources:
        # TODO: Refresh before scan?
        connection, table_name = data_source.split('.', 1)
        for column in Scanner(company_slug, connection).scan_columns(table_filter=table_name):
            data_reference = column['data_reference']
            try:
                for error in errors_by_column[data_source, data_reference]:
                    # Create field for each dataset that is missing the field
                    fields.append(
                        map_column_to_field(
                            column,
                            slug=error.field_slug,
                            data_source=error.dataset_slug,
                            is_identifier=error.identifier,
                        )
                    )
                del errors_by_column[data_source, data_reference]
            except KeyError:
                pass  # Column does not map to missing field

    # errors for fields we were not able to scan
    fields.extend(map_error_to_field(error) for error in itertools.chain.from_iterable(errors_by_column.values()))

    return fields



import re

from setuptools import find_namespace_packages, setup

with open("src/panoramic/cli/__version__.py", encoding="utf8") as f:
    version = re.search(r'__version__ = "(.*?)"', f.read()).group(1)  # type: ignore

TEST_REQUIRES = ["pytest>=5.3.5", "responses>=0.10.14", "freezegun>=0.3.15", "pytest-recording>=0.8.1"]
DEV_REQUIRES = ["mypy>=0.790", "flake8>=3.8.3", "black==20.8b0", "pre-commit>=2.1.1"]

setup(
    name="panoramic-cli",
    description="Panoramic Command Line Tool",
    url="https://github.com/panoramichq/panoramic-cli",
    project_urls={"Source Code": "https://github.com/panoramichq/panoramic-cli"},
    author="Panoramic",
    maintainer="Panoramic",
    version=version,
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    packages=find_namespace_packages(where='src', include=["panoramic.*"]),
    package_dir={"": "src"},
    python_requires=">=3.6",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "License :: OSI Approved",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
    ],
    install_requires=[
        'panoramic-auth>=0.2.0',
        'click>=7.1.2',
        'colorama>=0.4.3',
        'PyYAML==5.3.1',
        'packaging>=20.4',
        'tqdm>=4.47.0.',
        'python-dotenv>=0.14.0',
        'jsonschema>=3.0',
        'requests>=2.18.0',
        "importlib_resources ; python_version<'3.7'",
        "dbt>=0.18.1",
        "analytics-python==1.2.9",
    ],
    extras_require={"tests": TEST_REQUIRES, "dev": TEST_REQUIRES + DEV_REQUIRES},
    include_package_data=True,
    entry_points={"console_scripts": ["pano=panoramic.cli:cli"]},
)



import logging
from pathlib import Path
from typing import Optional

from panoramic.cli.file_utils import delete_file, write_yaml
from panoramic.cli.pano_model import (
    Actionable,
    PanoField,
    PanoModel,
    PanoVirtualDataSource,
)
from panoramic.cli.paths import FileExtension, Paths, PresetFileName

logger = logging.getLogger(__name__)


class FileWriter:
    """Responsible for writing data to local filesystem."""

    cwd: Path

    def __init__(self, *, cwd: Optional[Path] = None):
        if cwd is None:
            cwd = Path.cwd()

        # TODO: bunch of methods below ignore self.cwd
        self.cwd = cwd

    def delete(self, actionable: Actionable):
        """Delete data from local filesystem."""
        if isinstance(actionable, PanoModel):
            return self.delete_model(actionable)
        elif isinstance(actionable, PanoVirtualDataSource):
            return self.delete_data_source(actionable)
        elif isinstance(actionable, PanoField):
            return self.delete_field(actionable)
        else:
            raise NotImplementedError(f'write not implemented for type {type(actionable)}')

    def write(self, actionable: Actionable, *, package: Optional[str] = None, file_name: Optional[str] = None):
        """Write data to local filesystem."""
        if isinstance(actionable, PanoModel):
            return self.write_model(actionable, package=package, file_name=file_name)
        elif isinstance(actionable, PanoVirtualDataSource):
            return self.write_data_source(actionable, package=package)
        elif isinstance(actionable, PanoField):
            return self.write_field(actionable, package=package, file_name=file_name)
        else:
            raise NotImplementedError(f'write not implemented for type {type(actionable)}')

    def write_data_source(self, data_source: PanoVirtualDataSource, *, package: Optional[str] = None):
        """Write data source to local filesystem."""
        # Default to name of slugified name of DS
        package = package if package is not None else data_source.dataset_slug
        path = self.cwd / package / PresetFileName.DATASET_YAML.value
        logger.debug(f'About to write data source {data_source.id}')
        write_yaml(path, data_source.to_dict())

    def delete_data_source(self, data_source: PanoVirtualDataSource):
        """Delete data source from local filesystem."""
        assert data_source.package is not None
        path = self.cwd / data_source.package / PresetFileName.DATASET_YAML.value
        logger.debug(f'About to delete data source {data_source.id}')
        delete_file(path)

    def write_empty_model(self, model_name: str):
        """Create an empty model file."""
        path = Paths.scanned_dir() / f'{model_name}{FileExtension.MODEL_YAML.value}'
        logger.debug(f'About to create an empty model {model_name}')
        path.touch()

    def write_empty_field(self, field_slug: str):
        """Create an empty field file."""
        path = Paths.scanned_fields_dir() / f'{field_slug}{FileExtension.FIELD_YAML.value}'
        logger.debug(f'About to create an empty field {field_slug}')
        path.touch()

    def write_scanned_model(self, model: PanoModel):
        """Write scanned model to local filesystem."""
        path = Paths.scanned_dir() / f'{model.model_name}{FileExtension.MODEL_YAML.value}'
        logger.debug(f'About to write model {model.id}')
        write_yaml(path, model.to_dict())

    def write_scanned_field(self, field: PanoField):
        """"Write scanned field to local filesystem."""
        path = Paths.scanned_fields_dir() / f'{field.slug}{FileExtension.FIELD_YAML.value}'
        logger.debug(f'About to write field {field.slug}')
        write_yaml(path, field.to_dict())

    def write_model(self, model: PanoModel, *, package: Optional[str] = None, file_name: Optional[str] = None):
        """Write model to local filesystem."""
        # Default to name of slugified name of DS
        package_name = model.virtual_data_source if package is None else package
        assert package_name is not None  # TODO: virtual_data_source is Optional but shouldn't be
        if file_name is None:
            file_name = f'{model.model_name}{FileExtension.MODEL_YAML.value}'
        path = self.cwd / package_name / file_name
        logger.debug(f'About to write model {model.id}')
        write_yaml(path, model.to_dict())

    def delete_model(self, model: PanoModel):
        """Delete model from local filesystem."""
        assert model.package is not None
        assert model.file_name is not None
        path = self.cwd / model.package / model.file_name
        logger.debug(f'About to delete model {model.id}')
        delete_file(path)

    def write_field(self, field: PanoField, *, package: Optional[str] = None, file_name: Optional[str] = None):
        """Write model to local filesystem."""
        if file_name is None:
            file_name = f'{field.slug}{FileExtension.FIELD_YAML.value}'

        package = package if package is not None else field.data_source
        if package is not None:
            # dataset-scope field
            path = Paths.fields_dir(self.cwd / package) / file_name
        else:
            # company-scope field
            path = Paths.fields_dir(self.cwd) / file_name

        logger.debug(f'About to write field {field.id}')
        write_yaml(path, field.to_dict())

    def delete_field(self, field: PanoField):
        """Delete field from local filesystem."""
        assert field.file_name is not None

        if field.package is not None:
            # dataset-scope field
            path = Paths.fields_dir(self.cwd / field.package) / field.file_name
        else:
            # company-scope field
            path = Paths.fields_dir(self.cwd) / field.file_name

        logger.debug(f'About to delete field {field.id}')
        delete_file(path)



import logging
import os
from typing import Optional

import requests
from packaging import version
from tqdm import tqdm

from panoramic.cli.print import echo_error

URL = "https://a1.panocdn.com/updates/pano-cli/versions.json"

logger = logging.getLogger(__name__)


def fetch_analytics_write_key() -> Optional[str]:
    """We can reuse existing versions.json for purpose of distributing analytics write key.
    Later we can migrate to a different solution."""
    response = requests.get(URL, headers={'User-Agent': 'pano-cli/analytics'}, timeout=5)
    response.raise_for_status()
    data = response.json()
    return data.get('analytics_write_key')


def __fetch_minimum_supported_version(current_version) -> str:
    """Fetch minimum supported version of pano-cli from remote server.
    Returns JSON containing info about versions.
    Example:
    {
        "minimum_supported_version": "0.1.0",
        "latest_version": "0.1.0"
    }"""
    response = requests.get(URL, headers={"User-Agent": f"pano-cli/{current_version}"}, timeout=5)
    response.raise_for_status()
    data = response.json()
    return data['minimum_supported_version']


def __get_upgrade_command() -> str:
    """
    Read environment variables to understand how the CLI was installed.
    Return the appropriate command for the user to upgrade to the newest version.
    """
    if os.environ.get('RUNNING_UNDER_HOMEBREW') is not None:
        return "`brew update && brew upgrade panoramic-cli`"
    # Always fall back to assuming PIP
    return "`pip install --upgrade panoramic-cli`"


def is_version_supported(current_version: str) -> bool:
    """Check if current version of the CLI is still supported.
    If version has been deprecated print warning message notifying user to update the CLI.
    Returns bool. If check was successful program can continue otherwise it should be stopped.
    """
    try:
        minimum_supported_version = __fetch_minimum_supported_version(current_version)
    except requests.exceptions.RequestException:
        error_msg = 'Failed to connect to remote server to verify minimum supported CLI version.'
        echo_error(error_msg)
        logger.debug(error_msg, exc_info=True)
        return False
    except (KeyError, TypeError):
        error_msg = 'Failed to verify minimum supported CLI version.'
        echo_error(error_msg)
        logger.debug(error_msg, exc_info=True)
        return False

    if version.parse(current_version) < version.parse(minimum_supported_version):
        upgrade_command = __get_upgrade_command()
        message = (
            f"WARNING: This version '{current_version}' has been deprecated.\n"
            f"Please update to version '{minimum_supported_version}' or higher.\n"
            f"To update run: {upgrade_command}.\n"
        )
        tqdm.write(message)
        return False
    return True
